import ast
import logging
from _ast import AST

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import WordCompleter
from sqlalchemy.orm import DeclarativeBase as SQLTable, Session
from typing_extensions import Any, List, Optional, Tuple, Dict, Union, Type

from .datastructures import Case, PromptFor, CallableExpression, create_case, parse_string_to_expression


def prompt_user_for_expression(case: Union[Case, SQLTable], prompt_for: PromptFor, target_name: str,
                               output_type: Type, session: Optional[Session] = None) -> Tuple[str, CallableExpression]:
    """
    Prompt the user for an executable python expression.

    :param case: The case to classify.
    :param prompt_for: The type of information ask user about.
    :param target_name: The name of the target attribute to compare the case with.
    :param output_type: The type of the output of the given statement from the user.
    :param session: The sqlalchemy orm session.
    :return: A callable expression that takes a case and executes user expression on it.
    """
    while True:
        user_input, expression_tree = prompt_user_about_case(case, prompt_for, target_name)
        callable_expression = CallableExpression(user_input, output_type, expression_tree=expression_tree, session=session)
        try:
            callable_expression(case)
            break
        except Exception as e:
            logging.error(e)
            print(e)
    return user_input, callable_expression


def prompt_user_about_case(case: Union[Case, SQLTable], prompt_for: PromptFor, target_name: str) \
        -> Tuple[str, AST]:
    """
    Prompt the user for input.

    :param case: The case to prompt the user on.
    :param prompt_for: The type of information the user should provide for the given case.
    :param target_name: The name of the target property of the case that is queried.
    :return: The user input, and the executable expression that was parsed from the user input.
    """
    prompt_str = f"Give {prompt_for} for {case.__class__.__name__}.{target_name}"
    session = get_prompt_session_for_obj(case)
    user_input, expression_tree = prompt_user_input_and_parse_to_expression(prompt_str, session)
    return user_input, expression_tree


def get_completions(obj: Any) -> List[str]:
    """
    Get all completions for the object. This is used in the python prompt shell to provide completions for the user.

    :param obj: The object to get completions for.
    :return: A list of completions.
    """
    # Define completer with all object attributes and comparison operators
    completions = ['==', '!=', '>', '<', '>=', '<=', 'in', 'not', 'and', 'or', 'is']
    completions += ["isinstance(", "issubclass(", "type(", "len(", "hasattr(", "getattr(", "setattr(", "delattr("]
    completions += list(create_case(obj).keys())
    return completions


def prompt_user_input_and_parse_to_expression(prompt: Optional[str] = None, session: Optional[PromptSession] = None,
                                              user_input: Optional[str] = None) -> Tuple[str, ast.AST]:
    """
    Prompt the user for input.

    :param prompt: The prompt to display to the user.
    :param session: The prompt session to use.
    :param user_input: The user input to use. If given, the user input will be used instead of prompting the user.
    :return: The user input and the AST tree.
    """
    while True:
        if not user_input:
            user_input = session.prompt(f"\n{prompt} >>> ")
        if user_input.lower() in ['exit', 'quit', '']:
            break
        try:
            return user_input, parse_string_to_expression(user_input)
        except Exception as e:
            msg = f"Error parsing expression: {e}"
            logging.error(msg)
            print(msg)
            user_input = None


def get_prompt_session_for_obj(obj: Any) -> PromptSession:
    """
    Get a prompt session for an object.

    :param obj: The object to get the prompt session for.
    :return: The prompt session.
    """
    completions = get_completions(obj)
    completer = WordCompleter(completions)
    session = PromptSession(completer=completer)
    return session



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum

from anytree import NodeMixin
from typing_extensions import List, Optional, Self, Union, Dict, Any

from .datastructures import CallableExpression, Case, SQLTable
from .datastructures.enums import RDREdge, Stop
from .utils import SubclassJSONSerializer


class Rule(NodeMixin, SubclassJSONSerializer, ABC):
    fired: Optional[bool] = None
    """
    Whether the rule has fired or not.
    """

    def __init__(self, conditions: Optional[CallableExpression] = None,
                 conclusion: Optional[CallableExpression] = None,
                 parent: Optional[Rule] = None,
                 corner_case: Optional[Union[Case, SQLTable]] = None,
                 weight: Optional[str] = None):
        """
        A rule in the ripple down rules classifier.

        :param conditions: The conditions of the rule.
        :param conclusion: The conclusion of the rule when the conditions are met.
        :param parent: The parent rule of this rule.
        :param corner_case: The corner case that this rule is based on/created from.
        :param weight: The weight of the rule, which is the type of edge connecting the rule to its parent.
        """
        super(Rule, self).__init__()
        self.conclusion = conclusion
        self.corner_case = corner_case
        self.parent = parent
        self.weight: Optional[str] = weight
        self.conditions = conditions if conditions else None
        self.json_serialization: Optional[Dict[str, Any]] = None

    def _post_detach(self, parent):
        """
        Called after this node is detached from the tree, useful when drawing the tree.

        :param parent: The parent node from which this node was detached.
        """
        self.weight = None

    def __call__(self, x: Case) -> Self:
        return self.evaluate(x)

    def evaluate(self, x: Case) -> Rule:
        """
        Check if the rule or its refinement or its alternative match the case,
        by checking if the conditions are met, then return the rule that matches the case.

        :param x: The case to evaluate the rule on.
        :return: The rule that fired or the last evaluated rule if no rule fired.
        """
        if not self.conditions:
            raise ValueError("Rule has no conditions")
        self.fired = self.conditions(x)
        return self.evaluate_next_rule(x)

    @abstractmethod
    def evaluate_next_rule(self, x: Case):
        """
        Evaluate the next rule after this rule is evaluated.
        """
        pass

    @property
    def name(self):
        """
        Get the name of the rule, which is the conditions and the conclusion.
        """
        return self.__str__()

    def __str__(self, sep="\n"):
        """
        Get the string representation of the rule, which is the conditions and the conclusion.
        """
        return f"{self.conditions}{sep}=> {self.conclusion}"

    def __repr__(self):
        return self.__str__()


class HasAlternativeRule:
    """
    A mixin class for rules that have an alternative rule.
    """
    _alternative: Optional[Rule] = None
    """
    The alternative rule of the rule, which is evaluated when the rule doesn't fire.
    """
    furthest_alternative: Optional[List[Rule]] = None
    """
    The furthest alternative rule of the rule, which is the last alternative rule in the chain of alternative rules.
    """
    all_alternatives: Optional[List[Rule]] = None
    """
    All alternative rules of the rule, which is all the alternative rules in the chain of alternative rules.
    """

    @property
    def alternative(self) -> Optional[Rule]:
        return self._alternative

    @alternative.setter
    def alternative(self, new_rule: Rule):
        """
        Set the alternative rule of the rule. It is important that no rules should be retracted or changed,
        only new rules should be added.
        """
        if new_rule is None:
            return
        if self.furthest_alternative:
            self.furthest_alternative[-1].alternative = new_rule
        else:
            new_rule.parent = self
            new_rule.weight = RDREdge.Alternative.value if not new_rule.weight else new_rule.weight
            self._alternative = new_rule
        self.furthest_alternative = [new_rule]


class HasRefinementRule:
    _refinement: Optional[HasAlternativeRule] = None
    """
    The refinement rule of the rule, which is evaluated when the rule fires.
    """

    @property
    def refinement(self) -> Optional[Rule]:
        return self._refinement

    @refinement.setter
    def refinement(self, new_rule: Rule):
        """
        Set the refinement rule of the rule. It is important that no rules should be retracted or changed,
        only new rules should be added.
        """
        if new_rule is None:
            return
        new_rule.top_rule = self
        if self.refinement:
            self.refinement.alternative = new_rule
        else:
            new_rule.parent = self
            new_rule.weight = RDREdge.Refinement.value
            self._refinement = new_rule


class SingleClassRule(Rule, HasAlternativeRule, HasRefinementRule):
    """
    A rule in the SingleClassRDR classifier, it can have a refinement or an alternative rule or both.
    """

    def evaluate_next_rule(self, x: Case) -> SingleClassRule:
        if self.fired:
            returned_rule = self.refinement(x) if self.refinement else self
        else:
            returned_rule = self.alternative(x) if self.alternative else self
        return returned_rule if returned_rule.fired else self

    def fit_rule(self, x: Case, target: CallableExpression, conditions: CallableExpression):
        new_rule = SingleClassRule(conditions, target,
                                   corner_case=x, parent=self)
        if self.fired:
            self.refinement = new_rule
        else:
            self.alternative = new_rule

    def write_conclusion_as_source_code(self, parent_indent: str = "") -> str:
        """
        Get the source code representation of the conclusion of the rule.

        :param parent_indent: The indentation of the parent rule.
        """
        if isinstance(self.conclusion, CallableExpression):
            conclusion = self.conclusion.parsed_user_input
        elif isinstance(self.conclusion, Enum):
            conclusion = str(self.conclusion)
        else:
            conclusion = self.conclusion
        return f"{parent_indent}{' ' * 4}return {conclusion}\n"

    def write_condition_as_source_code(self, parent_indent: str = "") -> str:
        """
        Get the source code representation of the conditions of the rule.

        :param parent_indent: The indentation of the parent rule.
        """
        if_clause = "elif" if self.weight == RDREdge.Alternative.value else "if"
        return f"{parent_indent}{if_clause} {self.conditions.parsed_user_input}:\n"

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {"conditions": self.conditions.to_json(),
                                   "conclusion": self.conclusion.to_json(),
                                   "parent": self.parent.json_serialization if self.parent else None,
                                   "corner_case": self.corner_case.to_json() if self.corner_case else None,
                                   "weight": self.weight,
                                   "refinement": self.refinement.to_json() if self.refinement is not None else None,
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> SingleClassRule:
        loaded_rule = cls(conditions=CallableExpression.from_json(data["conditions"]),
                          conclusion=CallableExpression.from_json(data["conclusion"]),
                          parent=SingleClassRule.from_json(data["parent"]),
                          corner_case=Case.from_json(data["corner_case"]),
                          weight=data["weight"])
        loaded_rule.refinement = SingleClassRule.from_json(data["refinement"])
        loaded_rule.alternative = SingleClassRule.from_json(data["alternative"])
        return loaded_rule


class MultiClassStopRule(Rule, HasAlternativeRule):
    """
    A rule in the MultiClassRDR classifier, it can have an alternative rule and a top rule,
    the conclusion of the rule is a Stop category meant to stop the parent conclusion from being made.
    """
    top_rule: Optional[MultiClassTopRule] = None
    """
    The top rule of the rule, which is the nearest ancestor that fired and this rule is a refinement of.
    """

    def __init__(self, *args, **kwargs):
        super(MultiClassStopRule, self).__init__(*args, **kwargs)
        self.conclusion = Stop.stop

    def evaluate_next_rule(self, x: Case) -> Optional[Union[MultiClassStopRule, MultiClassTopRule]]:
        if self.fired:
            self.top_rule.fired = False
            return self.top_rule.alternative
        elif self.alternative:
            return self.alternative(x)
        else:
            return self.top_rule.alternative


class MultiClassTopRule(Rule, HasRefinementRule, HasAlternativeRule):
    """
    A rule in the MultiClassRDR classifier, it can have a refinement and a next rule.
    """

    def __init__(self, *args, **kwargs):
        super(MultiClassTopRule, self).__init__(*args, **kwargs)
        self.weight = RDREdge.Next.value

    def evaluate_next_rule(self, x: Case) -> Optional[Union[MultiClassStopRule, MultiClassTopRule]]:
        if self.fired and self.refinement:
            return self.refinement(x)
        elif self.alternative:  # Here alternative refers to next rule in MultiClassRDR
            return self.alternative

    def fit_rule(self, x: Case, target: CallableExpression, conditions: CallableExpression):
        if self.fired and target != self.conclusion:
            self.refinement = MultiClassStopRule(conditions, corner_case=x, parent=self)
        elif not self.fired:
            self.alternative = MultiClassTopRule(conditions, target, corner_case=x, parent=self)



from __future__ import annotations

import os
from unittest import TestCase

from sqlalchemy import ForeignKey
from sqlalchemy.orm import MappedAsDataclass, DeclarativeBase, declared_attr, Mapped, mapped_column, relationship
from typing_extensions import List, Any, Set

from ripple_down_rules.datastructures import CaseQuery, CallableExpression
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR
from ripple_down_rules.utils import render_tree


class Base(MappedAsDataclass, DeclarativeBase):

    @declared_attr.directive
    def __tablename__(cls) -> str:
        return cls.__name__


class PolymorphicIdentityMixin:

    @declared_attr.directive
    def __mapper_args__(cls):
        return {
            "polymorphic_identity": cls.__tablename__,
        }


class HasPart(Base):
    left_id: Mapped[int] = mapped_column(ForeignKey("PhysicalObject.id"), primary_key=True, init=False, repr=False)
    right_id: Mapped[int] = mapped_column(ForeignKey("PhysicalObject.id"), primary_key=True, init=False, repr=False)
    left: Mapped[PhysicalObject] = relationship(back_populates="has_part_relations", foreign_keys=[left_id])
    right: Mapped[PhysicalObject] = relationship(back_populates="part_of_relations", foreign_keys=[right_id])

    def __hash__(self):
        return hash(id(self))


class ContainsObject(Base):
    left_id: Mapped[int] = mapped_column(ForeignKey("PhysicalObject.id"), primary_key=True, init=False, repr=False)
    right_id: Mapped[int] = mapped_column(ForeignKey("PhysicalObject.id"), primary_key=True, init=False, repr=False)
    left: Mapped[PhysicalObject] = relationship(back_populates="contains_objects_relations", foreign_keys=[left_id])
    right: Mapped[PhysicalObject] = relationship(back_populates="is_contained_in_relations", foreign_keys=[right_id])

    def __hash__(self):
        return hash(id(self))


class PhysicalObject(Base):
    id: Mapped[int] = mapped_column(primary_key=True, init=False, autoincrement=True)
    name: Mapped[str]
    has_part_relations: Mapped[List[HasPart]] = relationship(init=False, back_populates="left",
                                                             foreign_keys=[HasPart.left_id],
                                                             repr=False, default_factory=list)
    part_of_relations: Mapped[List[HasPart]] = relationship(init=False, back_populates="right",
                                                            foreign_keys=[HasPart.right_id],
                                                            repr=False, default_factory=list)
    contains_objects_relations: Mapped[Set[ContainsObject]] = relationship(init=False, back_populates="left",
                                                                           foreign_keys=[ContainsObject.left_id],
                                                                           repr=False, default_factory=set)
    is_contained_in_relations: Mapped[Set[ContainsObject]] = relationship(init=False, back_populates="right",
                                                                          foreign_keys=[ContainsObject.right_id],
                                                                          repr=False, default_factory=set)
    type: Mapped[str] = mapped_column(init=False)

    @property
    def parts(self) -> List[PhysicalObject]:
        return [has_part.right for has_part in self.has_part_relations]

    @property
    def part_of(self) -> List[PhysicalObject]:
        return [has_part.left for has_part in self.part_of_relations]

    @property
    def contained_objects(self) -> Set[PhysicalObject]:
        return {cont_obj.right for cont_obj in self.contains_objects_relations}

    @property
    def is_contained_in(self) -> Set[PhysicalObject]:
        return {cont_obj.left for cont_obj in self.is_contained_in_relations}

    @declared_attr.directive
    def __mapper_args__(cls):
        return {
            "polymorphic_on": "type",
            "polymorphic_identity": cls.__tablename__,
        }

    def __hash__(self):
        return hash(id(self))


class RelationalRDRTestCase(TestCase):
    case: Any
    target: Any
    test_results_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"
    robot: PhysicalObject
    part_a: PhysicalObject
    part_b: PhysicalObject
    part_c: PhysicalObject
    part_d: PhysicalObject
    part_e: PhysicalObject
    part_f: PhysicalObject
    rob_has_parts: List[HasPart]
    containments: List[ContainsObject]

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(cls.test_results_dir):
            os.makedirs(cls.test_results_dir)
        cls.part_a = PhysicalObject(name="A")
        cls.part_b = PhysicalObject(name="B")
        cls.part_c = PhysicalObject(name="C")
        cls.part_d = PhysicalObject(name="D")
        cls.part_e = PhysicalObject(name="E")
        cls.part_f = PhysicalObject(name="F")
        robot = PhysicalObject(name="pr2")
        rob_parts = [cls.part_a, cls.part_b, cls.part_c, cls.part_d]
        cls.rob_has_parts = [HasPart(left=robot, right=part) for part in rob_parts]
        cls.containments = []
        cls.containments.append(ContainsObject(left=cls.part_a, right=cls.part_b))
        cls.containments.append(ContainsObject(left=cls.part_a, right=cls.part_c))
        cls.containments.append(ContainsObject(left=cls.part_c, right=cls.part_d))
        cls.containments.append(ContainsObject(left=cls.part_d, right=cls.part_e))
        cls.containments.append(ContainsObject(left=cls.part_e, right=cls.part_f))
        cls.robot: PhysicalObject = robot
        cls.target = CaseQuery(robot, robot.contained_objects,
                               {cls.part_b, cls.part_c, cls.part_d, cls.part_e})

    def test_setup(self):
        assert self.robot.parts == [self.part_a, self.part_b, self.part_c, self.part_d]
        assert all(len(part.part_of) == 1 and part.part_of[0] == self.robot for part in self.robot.parts)
        assert self.robot.contained_objects == set()
        assert self.part_a.contained_objects == {self.part_b, self.part_c}
        assert self.part_c.contained_objects == {self.part_d}
        assert self.part_d.contained_objects == {self.part_e}
        assert self.part_e.contained_objects == {self.part_f}

    def test_classify_scrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/relational_scrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        scrdr = SingleClassRDR()
        cat = scrdr.fit_case(CaseQuery(self.robot, self.robot.contained_objects), expert=expert)
        render_tree(scrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + "/relational_scrdr_classify")
        assert cat(self.robot) == self.target.target

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_parse_relational_conditions(self):
        user_input = "parts is not None and len(parts) > 0"
        conditions = CallableExpression(user_input, bool)
        print(conditions)
        print(conditions(self.robot))
        assert conditions(self.robot) == (self.robot.parts is not None and len(self.robot.parts) > 0)

    def test_parse_relational_conclusions(self):
        user_input = "parts.contained_objects"
        conclusion = CallableExpression(user_input, set)
        print(conclusion)
        print(conclusion(self.robot))
        assert conclusion(self.robot) == self.target.target



from __future__ import annotations

from abc import ABC, abstractmethod
from copy import copy, deepcopy

from matplotlib import pyplot as plt
from ordered_set import OrderedSet
from sqlalchemy.orm import DeclarativeBase as SQLTable, Session
from typing_extensions import List, Optional, Dict, Type, Union, Any, Self, Tuple

from .datastructures import Case, MCRDRMode, CallableExpression, CaseAttribute, CaseQuery
from .experts import Expert, Human
from .rules import Rule, SingleClassRule, MultiClassTopRule
from .utils import draw_tree, make_set, get_attribute_by_type, copy_case, \
    get_hint_for_attribute, SubclassJSONSerializer, is_iterable, make_list


class RippleDownRules(ABC):
    """
    The abstract base class for the ripple down rules classifiers.
    """
    fig: Optional[plt.Figure] = None
    """
    The figure to draw the tree on.
    """
    expert_accepted_conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the expert has accepted, such that they are not asked again.
    """

    def __init__(self, start_rule: Optional[Rule] = None, session: Optional[Session] = None):
        """
        :param start_rule: The starting rule for the classifier.
        :param session: The sqlalchemy orm session.
        """
        self.start_rule = start_rule
        self.session = session
        self.fig: Optional[plt.Figure] = None

    def __call__(self, case: Union[Case, SQLTable]) -> CaseAttribute:
        return self.classify(case)

    @abstractmethod
    def classify(self, case: Union[Case, SQLTable]) -> Optional[CaseAttribute]:
        """
        Classify a case.

        :param case: The case to classify.
        :return: The category that the case belongs to.
        """
        pass

    @abstractmethod
    def fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression]:
        """
        Fit the RDR on a case, and ask the expert for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        pass

    def fit(self, case_queries: List[CaseQuery],
            expert: Optional[Expert] = None,
            n_iter: int = None,
            animate_tree: bool = False,
            **kwargs_for_fit_case):
        """
        Fit the classifier to a batch of cases and categories.

        :param case_queries: The cases and categories to fit the classifier to.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param n_iter: The number of iterations to fit the classifier for.
        :param animate_tree: Whether to draw the tree while fitting the classifier.
        :param kwargs_for_fit_case: The keyword arguments to pass to the fit_case method.
        """
        cases = [case_query.case for case_query in case_queries]
        targets = [case.target for case in case_queries]
        if animate_tree:
            plt.ion()
        i = 0
        stop_iterating = False
        num_rules: int = 0
        while not stop_iterating:
            all_pred = 0
            if not targets:
                targets = [None] * len(cases)
            for case_query in case_queries:
                case = case_query.case
                target = case_query.target
                if not target:
                    conclusions = self.classify(case) if self.start_rule and self.start_rule.conditions else []
                    target = expert.ask_for_conclusion(case_query, conclusions)
                pred_cat = self.fit_case(case_query, expert=expert, **kwargs_for_fit_case)
                match = self.is_matching(pred_cat, target)
                if not match:
                    print(f"Predicted: {pred_cat} but expected: {target}")
                all_pred += int(match)
                if animate_tree and self.start_rule.size > num_rules:
                    num_rules = self.start_rule.size
                    self.update_figures()
            i += 1
            all_predictions = [1 if self.is_matching(self.classify(case), target) else 0
                               for case, target in zip(cases, targets)]
            all_pred = sum(all_predictions)
            print(f"Accuracy: {all_pred}/{len(targets)}")
            all_predicted = targets and all_pred == len(targets)
            num_iter_reached = n_iter and i >= n_iter
            stop_iterating = all_predicted or num_iter_reached
            if stop_iterating:
                break
        print(f"Finished training in {i} iterations")
        if animate_tree:
            plt.ioff()
            plt.show()

    @staticmethod
    def calculate_precision_and_recall(pred_cat: List[CaseAttribute], target: List[CaseAttribute]) -> Tuple[List[bool], List[bool]]:
        """
        :param pred_cat: The predicted category.
        :param target: The target category.
        :return: The precision and recall of the classifier.
        """
        pred_cat = pred_cat if is_iterable(pred_cat) else [pred_cat]
        target = target if is_iterable(target) else [target]
        recall = [not yi or (yi in pred_cat) for yi in target]
        target_types = [type(yi) for yi in target]
        precision = [(pred in target) or (type(pred) not in target_types) for pred in pred_cat]
        return precision, recall

    def is_matching(self, pred_cat: List[CaseAttribute], target: List[CaseAttribute]) -> bool:
        """
        :param pred_cat: The predicted category.
        :param target: The target category.
        :return: Whether the classifier is matching or not.
        """
        precision, recall = self.calculate_precision_and_recall(pred_cat, target)
        return all(recall) and all(precision)

    def update_figures(self):
        """
        Update the figures of the classifier.
        """
        if isinstance(self, GeneralRDR):
            for i, (_type, rdr) in enumerate(self.start_rules_dict.items()):
                if not rdr.fig:
                    rdr.fig = plt.figure(f"Rule {i}: {_type.__name__}")
                draw_tree(rdr.start_rule, rdr.fig)
        else:
            if not self.fig:
                self.fig = plt.figure(0)
            draw_tree(self.start_rule, self.fig)

    @staticmethod
    def case_has_conclusion(case: Union[Case, SQLTable], conclusion_type: Type) -> bool:
        """
        Check if the case has a conclusion.

        :param case: The case to check.
        :param conclusion_type: The target category type to compare the case with.
        :return: Whether the case has a conclusion or not.
        """
        if isinstance(case, SQLTable):
            prop_name, prop_value = get_attribute_by_type(case, conclusion_type)
            if hasattr(prop_value, "__iter__") and not isinstance(prop_value, str):
                return len(prop_value) > 0
            else:
                return prop_value is not None
        else:
            return conclusion_type in case


RDR = RippleDownRules


class SingleClassRDR(RippleDownRules, SubclassJSONSerializer):

    def fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression]:
        """
        Classify a case, and ask the user for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category if provided.

        :param case_query: The case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        expert = expert if expert else Human(session=self.session)
        case, attribute = case_query.case, case_query.attribute
        if case_query.target is None:
            target = expert.ask_for_conclusion(case_query)
        else:
            target = case_query.target

        if not self.start_rule:
            conditions = expert.ask_for_conditions(case, [target])
            self.start_rule = SingleClassRule(conditions, target, corner_case=case)

        pred = self.evaluate(case)

        if pred.conclusion != target:
            conditions = expert.ask_for_conditions(case, [target], pred)
            pred.fit_rule(case, target, conditions=conditions)

        return self.classify(case)

    def classify(self, case: Case) -> Optional[CaseAttribute]:
        """
        Classify a case by recursively evaluating the rules until a rule fires or the last rule is reached.
        """
        pred = self.evaluate(case)
        return pred.conclusion if pred.fired else None

    def evaluate(self, case: Case) -> SingleClassRule:
        """
        Evaluate the starting rule on a case.
        """
        matched_rule = self.start_rule(case)
        return matched_rule if matched_rule else self.start_rule

    def write_to_python_file(self, filename: str):
        """
        Write the tree of rules as source code to a file.
        """
        case = self.start_rule.corner_case
        if isinstance(case, Case):
            case_type = case._type
        else:
            case_type = type(case)
        case_module = case_type.__module__
        conclusion = self.start_rule.conclusion
        if isinstance(conclusion, CallableExpression):
            conclusion_types = [conclusion.conclusion_type]
        elif isinstance(conclusion, CaseAttribute):
            conclusion_types = list(conclusion._value_range)
        else:
            conclusion_types = [type(conclusion)]
        imports = ""
        if case_module != "builtins":
            imports += f"from {case_module} import {case_type.__name__}\n"
        if len(conclusion_types) > 1:
            conclusion_name = "Union[" + ", ".join([c.__name__ for c in conclusion_types]) + "]"
        else:
            conclusion_name = conclusion_types[0].__name__
        for conclusion_type in conclusion_types:
            if conclusion_type.__module__ != "builtins":
                imports += f"from {conclusion_type.__module__} import {conclusion_name}\n"
        imports += "from ripple_down_rules.datastructures import Case, create_case\n"
        imports += "\n\n"
        func_def = f"def classify_{conclusion_name.lower()}(case: {case_type.__name__}) -> {conclusion_name}:\n"
        with open(filename, "w") as f:
            f.write(imports)
            f.write(func_def)
            f.write(f"{' '*4}if not isinstance(case, Case):\n"
                    f"{' '*4}    case = create_case(case, recursion_idx=3)\n""")
            self.write_rules_as_source_code_to_file(self.start_rule, f, " " * 4)

    def write_rules_as_source_code_to_file(self, rule: SingleClassRule, file, parent_indent: str = ""):
        """
        Write the rules as source code to a file.
        """
        if rule.conditions:
            file.write(rule.write_condition_as_source_code(parent_indent))
            if rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, file, parent_indent + "    ")

            file.write(rule.write_conclusion_as_source_code(parent_indent))

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, file, parent_indent)

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rule": self.start_rule.to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create an instance of the class from a json
        """
        start_rule = SingleClassRule.from_json(data["start_rule"])
        return cls(start_rule)


class MultiClassRDR(RippleDownRules):
    """
    A multi class ripple down rules classifier, which can draw multiple conclusions for a case.
    This is done by going through all rules and checking if they fire or not, and adding stopping rules if needed,
    when wrong conclusions are made to stop these rules from firing again for similar cases.
    """
    evaluated_rules: Optional[List[Rule]] = None
    """
    The evaluated rules in the classifier for one case.
    """
    conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the case belongs to.
    """
    stop_rule_conditions: Optional[CallableExpression] = None
    """
    The conditions of the stopping rule if needed.
    """

    def __init__(self, start_rules: Optional[List[Rule]] = None,
                 mode: MCRDRMode = MCRDRMode.StopOnly, session: Optional[Session] = None):
        """
        :param start_rules: The starting rules for the classifier, these are the rules that are at the top of the tree
        and are always checked, in contrast to the refinement and alternative rules which are only checked if the
        starting rules fire or not.
        :param mode: The mode of the classifier, either StopOnly or StopPlusRule, or StopPlusRuleCombined.
        :param session: The sqlalchemy orm session.
        """
        self.start_rules = [MultiClassTopRule()] if not start_rules else start_rules
        super(MultiClassRDR, self).__init__(self.start_rules[0], session=session)
        self.mode: MCRDRMode = mode

    def classify(self, case: Union[Case, SQLTable]) -> List[Any]:
        evaluated_rule = self.start_rule
        self.conclusions = []
        while evaluated_rule:
            next_rule = evaluated_rule(case)
            if evaluated_rule.fired:
                self.add_conclusion(evaluated_rule)
            evaluated_rule = next_rule
        return self.conclusions

    def fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None,
                 add_extra_conclusions: bool = False) -> List[Union[CaseAttribute, CallableExpression]]:
        """
        Classify a case, and ask the user for stopping rules or classifying rules if the classification is incorrect
         or missing by comparing the case with the target category if provided.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions or for extra conclusions.
        :param add_extra_conclusions: Whether to add extra conclusions after classification is done.
        :return: The conclusions that the case belongs to.
        """
        expert = expert if expert else Human(session=self.session)
        case = case_query.case
        if case_query.target is None:
            targets = [expert.ask_for_conclusion(case_query)]
        else:
            targets = [case_query.target]
        self.expert_accepted_conclusions = []
        user_conclusions = []
        for target in targets:
            self.update_start_rule(case, target, expert)
            self.conclusions = []
            self.stop_rule_conditions = None
            evaluated_rule = self.start_rule
            while evaluated_rule:
                next_rule = evaluated_rule(case)
                good_conclusions = targets + user_conclusions + self.expert_accepted_conclusions

                if evaluated_rule.fired:
                    if target and evaluated_rule.conclusion not in good_conclusions:
                        # if self.case_has_conclusion(case, evaluated_rule.conclusion):
                        # Rule fired and conclusion is different from target
                        self.stop_wrong_conclusion_else_add_it(case, target, expert, evaluated_rule,
                                                               add_extra_conclusions)
                    else:
                        # Rule fired and target is correct or there is no target to compare
                        self.add_conclusion(evaluated_rule)

                if not next_rule:
                    if not make_set(target).intersection(make_set(self.conclusions)):
                        # Nothing fired and there is a target that should have been in the conclusions
                        self.add_rule_for_case(case, target, expert)
                        # Have to check all rules again to make sure only this new rule fires
                        next_rule = self.start_rule
                    elif add_extra_conclusions and not user_conclusions:
                        # No more conclusions can be made, ask the expert for extra conclusions if needed.
                        user_conclusions.extend(self.ask_expert_for_extra_conclusions(expert, case))
                        if user_conclusions:
                            next_rule = self.last_top_rule
                evaluated_rule = next_rule
        return self.conclusions

    def update_start_rule(self, case: Union[Case, SQLTable], target: Any, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case: The case to classify.
        :param target: The target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule.conditions:
            conditions = expert.ask_for_conditions(case, target)
            self.start_rule.conditions = conditions
            self.start_rule.conclusion = target
            self.start_rule.corner_case = case

    @property
    def last_top_rule(self) -> Optional[MultiClassTopRule]:
        """
        Get the last top rule in the tree.
        """
        if not self.start_rule.furthest_alternative:
            return self.start_rule
        else:
            return self.start_rule.furthest_alternative[-1]

    def stop_wrong_conclusion_else_add_it(self, case: Union[Case, SQLTable], target: Any, expert: Expert,
                                          evaluated_rule: MultiClassTopRule,
                                          add_extra_conclusions: bool):
        """
        Stop a wrong conclusion by adding a stopping rule.
        """
        if self.is_same_category_type(evaluated_rule.conclusion, target) \
                and self.is_conflicting_with_target(evaluated_rule.conclusion, target):
            self.stop_conclusion(case, target, expert, evaluated_rule)
        elif not self.conclusion_is_correct(case, target, expert, evaluated_rule, add_extra_conclusions):
            self.stop_conclusion(case, target, expert, evaluated_rule)

    def stop_conclusion(self, case: Union[Case, SQLTable], target: Any,
                        expert: Expert, evaluated_rule: MultiClassTopRule):
        """
        Stop a conclusion by adding a stopping rule.

        :param case: The case to classify.
        :param target: The target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param evaluated_rule: The evaluated rule to ask the expert about.
        """
        conditions = expert.ask_for_conditions(case, target, evaluated_rule)
        evaluated_rule.fit_rule(case, target, conditions=conditions)
        if self.mode == MCRDRMode.StopPlusRule:
            self.stop_rule_conditions = conditions
        if self.mode == MCRDRMode.StopPlusRuleCombined:
            new_top_rule_conditions = conditions.combine_with(evaluated_rule.conditions)
            self.add_top_rule(new_top_rule_conditions, target, case)

    @staticmethod
    def is_conflicting_with_target(conclusion: Any, target: Any) -> bool:
        """
        Check if the conclusion is conflicting with the target category.

        :param conclusion: The conclusion to check.
        :param target: The target category to compare the conclusion with.
        :return: Whether the conclusion is conflicting with the target category.
        """
        if hasattr(conclusion, "mutually_exclusive") and conclusion.mutually_exclusive:
            return True
        else:
            return not make_set(conclusion).issubset(make_set(target))

    @staticmethod
    def is_same_category_type(conclusion: Any, target: Any) -> bool:
        """
        Check if the conclusion is of the same class as the target category.

        :param conclusion: The conclusion to check.
        :param target: The target category to compare the conclusion with.
        :return: Whether the conclusion is of the same class as the target category but has a different value.
        """
        return conclusion.__class__ == target.__class__ and target.__class__ != CaseAttribute

    def conclusion_is_correct(self, case: Union[Case, SQLTable], target: Any, expert: Expert, evaluated_rule: Rule,
                              add_extra_conclusions: bool) -> bool:
        """
        Ask the expert if the conclusion is correct, and add it to the conclusions if it is.

        :param case: The case to classify.
        :param target: The target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param evaluated_rule: The evaluated rule to ask the expert about.
        :param add_extra_conclusions: Whether adding extra conclusions after classification is allowed.
        :return: Whether the conclusion is correct or not.
        """
        conclusions = list(OrderedSet(self.conclusions))
        if (add_extra_conclusions and expert.ask_if_conclusion_is_correct(case, evaluated_rule.conclusion,
                                                                          targets=target,
                                                                          current_conclusions=conclusions)):
            self.add_conclusion(evaluated_rule)
            self.expert_accepted_conclusions.append(evaluated_rule.conclusion)
            return True
        return False

    def add_rule_for_case(self, case: Union[Case, SQLTable], target: Any, expert: Expert):
        """
        Add a rule for a case that has not been classified with any conclusion.
        """
        if self.stop_rule_conditions and self.mode == MCRDRMode.StopPlusRule:
            conditions = self.stop_rule_conditions
            self.stop_rule_conditions = None
        else:
            conditions = expert.ask_for_conditions(case, target)
        self.add_top_rule(conditions, target, case)

    def ask_expert_for_extra_conclusions(self, expert: Expert, case: Union[Case, SQLTable]) -> List[Any]:
        """
        Ask the expert for extra conclusions when no more conclusions can be made.

        :param expert: The expert to ask for extra conclusions.
        :param case: The case to ask extra conclusions for.
        :return: The extra conclusions that the expert has provided.
        """
        extra_conclusions = []
        conclusions = list(OrderedSet(self.conclusions))
        if not expert.use_loaded_answers:
            print("current conclusions:", conclusions)
        extra_conclusions_dict = expert.ask_for_extra_conclusions(case, conclusions)
        if extra_conclusions_dict:
            for conclusion, conditions in extra_conclusions_dict.items():
                self.add_top_rule(conditions, conclusion, case)
                extra_conclusions.append(conclusion)
        return extra_conclusions

    def add_conclusion(self, evaluated_rule: Rule) -> None:
        """
        Add the conclusion of the evaluated rule to the list of conclusions.

        :param evaluated_rule: The evaluated rule to add the conclusion of.
        """
        conclusion_types = [type(c) for c in self.conclusions]
        if type(evaluated_rule.conclusion) not in conclusion_types:
            self.conclusions.extend(make_list(evaluated_rule.conclusion))
        else:
            same_type_conclusions = [c for c in self.conclusions if type(c) == type(evaluated_rule.conclusion)]
            combined_conclusion = evaluated_rule.conclusion if isinstance(evaluated_rule.conclusion, set) \
                else {evaluated_rule.conclusion}
            combined_conclusion = copy(combined_conclusion)
            for c in same_type_conclusions:
                combined_conclusion.update(c if isinstance(c, set) else make_set(c))
                self.conclusions.remove(c)
            self.conclusions.extend(make_list(combined_conclusion))

    def add_top_rule(self, conditions: CallableExpression, conclusion: Any, corner_case: Union[Case, SQLTable]):
        """
        Add a top rule to the classifier, which is a rule that is always checked and is part of the start_rules list.

        :param conditions: The conditions of the rule.
        :param conclusion: The conclusion of the rule.
        :param corner_case: The corner case of the rule.
        """
        self.start_rule.alternative = MultiClassTopRule(conditions, conclusion, corner_case=corner_case)


class GeneralRDR(RippleDownRules):
    """
    A general ripple down rules classifier, which can draw multiple conclusions for a case, but each conclusion is part
    of a set of mutually exclusive conclusions. Whenever a conclusion is made, the classification restarts from the
    starting rule, and all the rules that belong to the class of the made conclusion are not checked again. This
    continues until no more rules can be fired. In addition, previous conclusions can be used as conditions or input to
    the next classification/cycle.
    Another possible mode is to have rules that are considered final, when fired, inference will not be restarted,
     and only a refinement can be made to the final rule, those can also be used in another SCRDR of their own that
     gets called when the final rule fires.
    """

    def __init__(self, category_rdr_map: Optional[Dict[Type, Union[SingleClassRDR, MultiClassRDR]]] = None):
        """
        :param category_rdr_map: A map of categories to ripple down rules classifiers,
        where each category is a parent category that has a set of mutually exclusive (in case of SCRDR) child
        categories, e.g. {Species: SCRDR, Habitat: MCRDR}, where Species and Habitat are parent categories and SCRDR
        and MCRDR are SingleClass and MultiClass ripple down rules classifiers. Species can have child categories like
        Mammal, Bird, Fish, etc. which are mutually exclusive, and Habitat can have child categories like
        Land, Water, Air, etc, which are not mutually exclusive due to some animals living more than one habitat.
        """
        self.start_rules_dict: Dict[Type, Union[SingleClassRDR, MultiClassRDR]] \
            = category_rdr_map if category_rdr_map else {}
        super(GeneralRDR, self).__init__()
        self.all_figs: List[plt.Figure] = [sr.fig for sr in self.start_rules_dict.values()]

    @property
    def start_rule(self) -> Optional[Union[SingleClassRule, MultiClassTopRule]]:
        return self.start_rules[0] if self.start_rules_dict else None

    @start_rule.setter
    def start_rule(self, value: Union[SingleClassRDR, MultiClassRDR]):
        if value:
            self.start_rules_dict[type(value.start_rule.conclusion)] = value

    @property
    def start_rules(self) -> List[Union[SingleClassRule, MultiClassTopRule]]:
        return [rdr.start_rule for rdr in self.start_rules_dict.values()]

    def classify(self, case: Union[Case, SQLTable]) -> Optional[List[Any]]:
        """
        Classify a case by going through all RDRs and adding the categories that are classified, and then restarting
        the classification until no more categories can be added.

        :param case: The case to classify.
        :return: The categories that the case belongs to.
        """
        conclusions = []
        case_cp = copy_case(case)
        while True:
            added_attributes = False
            for cat_type, rdr in self.start_rules_dict.items():
                if self.case_has_conclusion(case_cp, cat_type):
                    continue
                pred_atts = rdr.classify(case_cp)
                if pred_atts:
                    pred_atts = make_list(pred_atts)
                    pred_atts = [p for p in pred_atts if p not in conclusions]
                    added_attributes = True
                    conclusions.extend(pred_atts)
                    GeneralRDR.update_case(case_cp, pred_atts)
            if not added_attributes:
                break
        return conclusions

    def fit_case(self, case_queries: List[CaseQuery], expert: Optional[Expert] = None, **kwargs) \
            -> List[Union[CaseAttribute, CallableExpression]]:
        """
        Fit the GRDR on a case, if the target is a new type of category, a new RDR is created for it,
        else the existing RDR of that type will be fitted on the case, and then classification is done and all
        concluded categories are returned. If the category is mutually exclusive, an SCRDR is created, else an MCRDR.
        In case of SCRDR, multiple conclusions of the same type replace each other, in case of MCRDR, they are added if
        they are accepted by the expert, and the attribute of that category is represented in the case as a set of
        values.

        :param case_queries: The queries containing the case to classify and the target categories to compare the case
        with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The categories that the case belongs to.
        """
        expert = expert if expert else Human()
        case_queries = [case_queries] if not isinstance(case_queries, list) else case_queries
        assert len(case_queries) > 0, "No case queries provided"
        case = case_queries[0].case
        assert all([case is case_query.case for case_query in case_queries]), ("fit_case requires only one case,"
                                                                               " for multiple cases use fit instead")
        case_query_cp = copy(case_queries[0])
        case_cp = case_query_cp.case
        for case_query in case_queries:
            target = case_query.target
            if not target:
                target = expert.ask_for_conclusion(case_query)
            case_query_cp = CaseQuery(case_cp, attribute_name=case_query.attribute_name, target=target)
            if is_iterable(target) and not isinstance(target, CaseAttribute):
                target_type = type(make_list(target)[0])
                assert all([type(t) is target_type for t in target]), ("All targets of a case query must be of the same"
                                                                       " type")
            else:
                target_type = type(target)
            if target_type not in self.start_rules_dict:
                conclusions = self.classify(case)
                self.update_case(case_cp, conclusions)
                new_rdr = self.initialize_new_rdr_for_attribute(target, case_cp)
                new_conclusions = new_rdr.fit_case(case_query_cp, expert, **kwargs)
                self.start_rules_dict[target_type] = new_rdr
                self.update_case(case_cp, new_conclusions, target_type)
            elif not self.case_has_conclusion(case_cp, target_type):
                for rdr_type, rdr in self.start_rules_dict.items():
                    if target_type is not rdr_type:
                        conclusions = rdr.classify(case_cp)
                    else:
                        conclusions = self.start_rules_dict[target_type].fit_case(case_query_cp,
                                                                                   expert, **kwargs)
                    self.update_case(case_cp, conclusions, rdr_type)

        return self.classify(case)

    @staticmethod
    def initialize_new_rdr_for_attribute(attribute: Any, case: Union[Case, SQLTable]):
        """
        Initialize the appropriate RDR type for the target.
        """
        if isinstance(case, SQLTable):
            prop = get_attribute_by_type(case, type(attribute))
            if hasattr(prop, "__iter__") and not isinstance(prop, str):
                return MultiClassRDR()
            else:
                return SingleClassRDR()
        elif isinstance(attribute, CaseAttribute):
            return SingleClassRDR() if attribute.mutually_exclusive else MultiClassRDR()
        else:
            return MultiClassRDR() if is_iterable(attribute) else SingleClassRDR()

    @staticmethod
    def update_case(case: Union[Case, SQLTable],
                    conclusions: List[Any], attribute_type: Optional[Any] = None):
        """
        Update the case with the conclusions.

        :param case: The case to update.
        :param conclusions: The conclusions to update the case with.
        :param attribute_type: The type of the attribute to update.
        """
        if not conclusions:
            return
        conclusions = [conclusions] if not isinstance(conclusions, list) else list(conclusions)
        if len(conclusions) == 0:
            return
        if isinstance(case, SQLTable):
            conclusions_type = type(conclusions[0]) if not attribute_type else attribute_type
            attr_name, attribute = get_attribute_by_type(case, conclusions_type)
            hint, origin, args = get_hint_for_attribute(attr_name, case)
            if isinstance(attribute, set) or origin == set:
                attribute = set() if attribute is None else attribute
                for c in conclusions:
                    attribute.update(make_set(c))
            elif isinstance(attribute, list) or origin == list:
                attribute = [] if attribute is None else attribute
                attribute.extend(conclusions)
            elif len(conclusions) == 1 and hint == conclusions_type:
                setattr(case, attr_name, conclusions.pop())
            else:
                raise ValueError(f"Cannot add multiple conclusions to attribute {attr_name}")
        else:
            for c in make_set(conclusions):
                case.update(c.as_dict)

    @property
    def names_of_all_types(self) -> List[str]:
        """
        Get the names of all the types of categories that the GRDR can classify.
        """
        return [t.__name__ for t in self.start_rules_dict.keys()]

    @property
    def all_types(self) -> List[Type]:
        """
        Get all the types of categories that the GRDR can classify.
        """
        return list(self.start_rules_dict.keys())



import os
from unittest import TestCase, skip

from typing_extensions import List

from ripple_down_rules.datasets import Habitat, Species
from ripple_down_rules.datasets import load_zoo_dataset
from ripple_down_rules.datastructures import Case, MCRDRMode, \
    Case, CaseAttribute, Category, CaseQuery
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR, MultiClassRDR, GeneralRDR
from ripple_down_rules.utils import render_tree, get_all_subclasses


class TestRDR(TestCase):
    all_cases: List[Case]
    targets: List[str]
    test_results_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"
    generated_rdrs_dir: str = "./test_generated_rdrs"
    cache_file: str = f"{test_results_dir}/zoo_dataset.pkl"

    @classmethod
    def setUpClass(cls):
        # fetch dataset
        cls.all_cases, cls.targets = load_zoo_dataset(cache_file=cls.cache_file)
        for test_dir in [cls.test_results_dir, cls.expert_answers_dir, cls.generated_rdrs_dir]:
            if not os.path.exists(test_dir):
                os.makedirs(test_dir)

    def test_classify_scrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/scrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        scrdr = SingleClassRDR()
        cat = scrdr.fit_case(CaseQuery(self.all_cases[0], target=self.targets[0]), expert=expert)
        self.assertEqual(cat, self.targets[0])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_scrdr(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        filename = self.expert_answers_dir + "/scrdr_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        scrdr = SingleClassRDR()
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        scrdr.fit(case_queries, expert=expert,
                  animate_tree=draw_tree)
        scrdr.write_to_python_file(self.generated_rdrs_dir + "/scrdr.py")
        render_tree(scrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/scrdr")

        cat = scrdr.classify(self.all_cases[50])
        self.assertEqual(cat, self.targets[50])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_classify_mcrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/mcrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        mcrdr = MultiClassRDR()
        cats = mcrdr.fit_case(CaseQuery(self.all_cases[0], target=self.targets[0]), expert=expert)

        self.assertEqual(cats[0], self.targets[0])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_only(self):
        use_loaded_answers = True
        draw_tree = False
        save_answers = False
        filename = self.expert_answers_dir + "/mcrdr_expert_answers_stop_only_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR()
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree, n_iter=1)
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_stop_only")
        cats = mcrdr.classify(self.all_cases[50])
        self.assertEqual(cats[0], self.targets[50])
        self.assertTrue(len(cats) == 1)
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_plus_rule(self):
        use_loaded_answers = True
        draw_tree = False
        save_answers = False
        append = False
        filename = self.expert_answers_dir + "/mcrdr_stop_plus_rule_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR(mode=MCRDRMode.StopPlusRule)
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        try:
            mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        # catch pop from empty list error
        except IndexError as e:
            if append:
                expert.use_loaded_answers = False
                mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
            else:
                raise e
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_stop_plus_rule")
        cats = mcrdr.classify(self.all_cases[50])
        self.assertEqual(cats[0], self.targets[50])
        self.assertTrue(len(cats) == 1)
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file, append=append)

    def test_fit_mcrdr_stop_plus_rule_combined(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        append = False
        filename = self.expert_answers_dir + "/mcrdr_stop_plus_rule_combined_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR(mode=MCRDRMode.StopPlusRuleCombined)
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        try:
            mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        # catch pop from empty list error
        except IndexError as e:
            if append:
                expert.use_loaded_answers = False
                mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
            else:
                raise e
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_stop_plus_rule_combined")
        cats = mcrdr.classify(self.all_cases[50])
        self.assertEqual(cats[0], self.targets[50])
        self.assertTrue(len(cats) == 1)
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file, append=append)

    @skip("Extra conclusions loaded answers are not working with new prompt interface")
    def test_classify_mcrdr_with_extra_conclusions(self):
        use_loaded_answers = True
        save_answers = False
        file_name = self.expert_answers_dir + "/mcrdr_extra_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(file_name)

        mcrdr = MultiClassRDR()
        cats = mcrdr.fit_case(CaseQuery(self.all_cases[0], target=self.targets[0]),
                              add_extra_conclusions=True, expert=expert)
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_extra_classify")
        LivesOnlyOnLand = get_all_subclasses(CaseAttribute)["LivesOnlyOnLand".lower()]
        self.assertEqual(cats, [self.targets[50], LivesOnlyOnLand(True)])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, file_name)
            expert.save_answers(file)

    @skip("Extra conclusions loaded answers are not working with new prompt interface")
    def test_fit_mcrdr_with_extra_conclusions(self):
        draw_tree = False
        use_loaded_answers = True
        save_answers = False
        expert = Human(use_loaded_answers=use_loaded_answers)
        mcrdr = MultiClassRDR()
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        mcrdr.fit(case_queries, add_extra_conclusions=False, expert=expert, animate_tree=False)
        expert = Human(use_loaded_answers=use_loaded_answers)
        file_name = self.expert_answers_dir + "/mcrdr_extra_expert_answers_fit"
        if use_loaded_answers:
            expert.load_answers(file_name)
        mcrdr.fit(case_queries, add_extra_conclusions=True, expert=expert, n_iter=10, animate_tree=draw_tree)
        cats = mcrdr.classify(self.all_cases[50])
        LivesOnlyOnLand = get_all_subclasses(CaseAttribute)["LivesOnlyOnLand".lower()]
        self.assertEqual(cats, [self.targets[50], LivesOnlyOnLand(True)])
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_extra")
        if save_answers:
            expert.save_answers(file_name)

    def test_classify_grdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/grdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        grdr = GeneralRDR()

        targets = [self.targets[0], Habitat.land]
        case_queries = [CaseQuery(self.all_cases[0], target=target) for target in targets]
        cats = grdr.fit_case(case_queries, expert=expert)
        self.assertEqual(cats, targets)

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_grdr(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        filename = self.expert_answers_dir + "/grdr_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        fit_scrdr = self.get_fit_scrdr(draw_tree=False)

        grdr = GeneralRDR({type(fit_scrdr.start_rule.conclusion): fit_scrdr})

        def get_habitat(x: Case, t: Category):
            all_habs = []
            if t == Species.mammal and x["aquatic"] == 0:
                all_habs.append({Habitat.land})
            elif t == Species.bird:
                all_habs.append({Habitat.land})
                if x["airborne"] == 1:
                    all_habs[-1].update({Habitat.air})
                if x["aquatic"] == 1:
                    all_habs[-1].update({Habitat.water})
            elif t == Species.fish:
                all_habs.append({Habitat.water})
            elif t == Species.molusc:
                all_habs.append({Habitat.land})
                if x["aquatic"] == 1:
                    all_habs[-1].update({Habitat.water})
            return all_habs + [t]

        n = 20
        habitat_targets = [get_habitat(x, t) for x, t in zip(self.all_cases[:n], self.targets[:n])]
        case_queries = [CaseQuery(case, target=target)
                        for case, targets in zip(self.all_cases[:n], habitat_targets)
                        for target in targets]
        grdr.fit(case_queries, expert=expert,
                 animate_tree=draw_tree, n_iter=n)
        for rule in grdr.start_rules:
            render_tree(rule, use_dot_exporter=True,
                        filename=self.test_results_dir + f"/grdr_{type(rule.conclusion).__name__}")

        cats = grdr.classify(self.all_cases[50])
        self.assertEqual(cats, [self.targets[50], Habitat.land])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    @skip("Extra conclusions loaded answers are not working with new prompt interface")
    def test_fit_grdr_with_extra_conclusions(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        filename = self.expert_answers_dir + "/grdr_expert_answers_fit_extra"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        fit_scrdr = self.get_fit_scrdr(draw_tree=False)

        grdr = GeneralRDR({type(fit_scrdr.start_rule.conclusion): fit_scrdr})

        grdr.fit(self.all_cases[17:], expert=expert,
                 animate_tree=draw_tree, n_iter=20, add_extra_conclusions=True)
        for rule in grdr.start_rules:
            render_tree(rule, use_dot_exporter=True,
                        filename=self.test_results_dir + f"/grdr_{type(rule.conclusion).__name__}")

        cats = grdr.classify(self.all_cases[50])
        self.assertEqual(cats, [self.targets[50], Habitat.land])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def get_fit_scrdr(self, draw_tree=False) -> SingleClassRDR:
        filename = self.expert_answers_dir + "/scrdr_expert_answers_fit"
        expert = Human(use_loaded_answers=True)
        expert.load_answers(filename)

        scrdr = SingleClassRDR()
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        scrdr.fit(case_queries, expert=expert,
                  animate_tree=draw_tree)
        return scrdr



from __future__ import annotations

from dataclasses import dataclass

from sqlalchemy.orm import DeclarativeBase as SQLTable
from typing_extensions import Any, Optional, Type

from .case import create_case, Case
from ..utils import get_attribute_name, copy_case


@dataclass
class CaseQuery:
    """
    This is a dataclass that represents an attribute of an object and its target value. If attribute name is
    not provided, it will be inferred from the attribute itself or from the attribute type or from the target value,
    depending on what is provided.
    """
    case: Any
    """
    The case that the attribute belongs to.
    """
    attribute: Optional[Any] = None
    """
    The attribute itself.
    """
    targets: Optional[Any] = None
    """
    The target value of the attribute.
    """
    attribute_name: Optional[str] = None
    """
    The name of the attribute.
    """
    attribute_type: Optional[Type] = None
    """
    The type of the attribute.
    """
    relational_representation: Optional[str] = None
    """
    The representation of the target value in relational form.
    """

    def __init__(self, case: Any, attribute: Optional[Any] = None, target: Optional[Any] = None,
                 attribute_name: Optional[str] = None, attribute_type: Optional[Type] = None,
                 relational_representation: Optional[str] = None):

        if attribute_name is None:
            attribute_name = get_attribute_name(case, attribute, attribute_type, target)
        self.attribute_name = attribute_name

        if not isinstance(case, (Case, SQLTable)):
            case = create_case(case, max_recursion_idx=3)
        self.case = case

        self.attribute = getattr(self.case, self.attribute_name) if self.attribute_name else None
        self.attribute_type = type(self.attribute) if self.attribute else None
        self.target = target
        self.relational_representation = relational_representation

    @property
    def name(self):
        return self.attribute_name if self.attribute_name else self.__class__.__name__

    def __str__(self):
        if self.relational_representation:
            return f"{self.name} |= {self.relational_representation}"
        else:
            return f"{self.target}"

    def __repr__(self):
        return self.__str__()

    def __copy__(self):
        return CaseQuery(copy_case(self.case), attribute_name=self.attribute_name, target=self.target)



from __future__ import annotations

import importlib
import json
import logging
from abc import abstractmethod
from collections import UserDict
from copy import deepcopy
from dataclasses import dataclass

import matplotlib
import networkx as nx
from anytree import Node, RenderTree
from anytree.exporter import DotExporter
from matplotlib import pyplot as plt
from sqlalchemy import MetaData, inspect
from sqlalchemy.orm import Mapped, registry, class_mapper, DeclarativeBase as SQLTable, Session
from tabulate import tabulate
from typing_extensions import Callable, Set, Any, Type, Dict, TYPE_CHECKING, get_type_hints, \
    get_origin, get_args, Tuple, Optional, List, Union, Self

if TYPE_CHECKING:
    from .datastructures import Case

matplotlib.use("Qt5Agg")  # or "Qt5Agg", depending on availability


def make_list(value: Any) -> List:
    """
    Make a list from a value.

    :param value: The value to make a list from.
    """
    return list(value) if is_iterable(value) else [value]


def is_iterable(obj: Any) -> bool:
    """
    Check if an object is iterable.

    :param obj: The object to check.
    """
    return hasattr(obj, "__iter__") and not isinstance(obj, (str, type))


def get_type_from_string(type_path: str):
    """
    Get a type from a string describing its path using the format "module_path.ClassName".

    :param type_path: The path to the type.
    """
    module_path, class_name = type_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


def get_full_class_name(cls):
    """
    Returns the full name of a class, including the module name.
    Copied from: https://github.com/tomsch420/random-events/blob/master/src/random_events/utils.py#L6C1-L21C101

    :param cls: The class.
    :return: The full name of the class
    """
    return cls.__module__ + "." + cls.__name__


def recursive_subclasses(cls):
    """
    Copied from: https://github.com/tomsch420/random-events/blob/master/src/random_events/utils.py#L6C1-L21C101
    :param cls: The class.
    :return: A list of the classes subclasses.
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in recursive_subclasses(s)]


class SubclassJSONSerializer:
    """
    Copied from: https://github.com/tomsch420/random-events/blob/master/src/random_events/utils.py#L6C1-L21C101
    Class for automatic (de)serialization of subclasses.
    Classes that inherit from this class can be serialized and deserialized automatically by calling this classes
    'from_json' method.
    """

    def to_json_file(self, filename: str):
        """
        Save the object to a json file.
        """
        data = self.to_json()
        # save the json to a file
        if not filename.endswith(".json"):
            filename += ".json"
        with open(filename, "w") as f:
            json.dump(data, f, indent=4)
        return data

    def to_json(self) -> Dict[str, Any]:
        return {"_type": get_full_class_name(self.__class__), **self._to_json()}

    @abstractmethod
    def _to_json(self) -> Dict[str, Any]:
        """
        Create a json dict from the object.
        """
        pass

    @classmethod
    @abstractmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create a variable from a json dict.
        This method is called from the from_json method after the correct subclass is determined and should be
        overwritten by the respective subclass.

        :param data: The json dict
        :return: The deserialized object
        """
        raise NotImplementedError()

    @classmethod
    def from_json_file(cls, filename: str):
        """
        Create an instance of the subclass from the data in the given json file.

        :param filename: The filename of the json file.
        """
        if not filename.endswith(".json"):
            filename += ".json"
        with open(filename, "r") as f:
            scrdr_json = json.load(f)
        return cls.from_json(scrdr_json)

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create the correct instanceof the subclass from a json dict.

        :param data: The json dict
        :return: The correct instance of the subclass
        """
        if data is None:
            return None
        if get_full_class_name(cls) == data["_type"]:
            return cls._from_json(data)
        for subclass in recursive_subclasses(SubclassJSONSerializer):
            if get_full_class_name(subclass) == data["_type"]:
                subclass_data = deepcopy(data)
                subclass_data.pop("_type")
                return subclass._from_json(subclass_data)

        raise ValueError("Unknown type {}".format(data["_type"]))

    save = to_json_file
    load = from_json_file


def copy_case(case: Union[Case, SQLTable]) -> Union[Case, SQLTable]:
    """
    Copy a case.

    :param case: The case to copy.
    :return: The copied case.
    """
    if isinstance(case, SQLTable):
        return copy_orm_instance_with_relationships(case)
    else:
        return deepcopy(case)


def copy_orm_instance(instance: SQLTable) -> SQLTable:
    """
    Copy an ORM instance by expunging it from the session then deep copying it and adding it back to the session. This
    is useful when you want to copy an instance and make changes to it without affecting the original instance.

    :param instance: The instance to copy.
    :return: The copied instance.
    """
    session: Session = inspect(instance).session
    session.expunge(instance)
    new_instance = deepcopy(instance)
    session.add(instance)
    return new_instance


def copy_orm_instance_with_relationships(instance: SQLTable) -> SQLTable:
    """
    Copy an ORM instance with its relationships (i.e. its foreign keys).

    :param instance: The instance to copy.
    :return: The copied instance.
    """
    instance_cp = copy_orm_instance(instance)
    for rel in class_mapper(instance.__class__).relationships:
        related_obj = getattr(instance, rel.key)
        if related_obj is not None:
            setattr(instance_cp, rel.key, related_obj)
    return instance_cp


def get_value_type_from_type_hint(attr_name: str, obj: Any) -> Type:
    """
    Get the value type from the type hint of an object attribute.

    :param attr_name: The name of the attribute.
    :param obj: The object to get the attributes from.
    """
    hint, origin, args = get_hint_for_attribute(attr_name, obj)
    if not origin and not hint:
        if hasattr(obj, attr_name):
            attr_value = getattr(obj, attr_name)
            if attr_value is not None:
                return type(attr_value)
        raise ValueError(f"Couldn't get type for Attribute {attr_name}, please provide a type hint")
    if origin in [list, set, tuple, type, dict]:
        attr_value_type = args[0]
    elif hint:
        attr_value_type = hint
    else:
        raise ValueError(f"Attribute {attr_name} has unsupported type {hint}.")
    return attr_value_type


def get_hint_for_attribute(attr_name: str, obj: Any) -> Tuple[Optional[Any], Optional[Any], Tuple[Any]]:
    """
    Get the type hint for an attribute of an object.

    :param attr_name: The name of the attribute.
    :param obj: The object to get the attribute from.
    :return: The type hint of the attribute.
    """
    if attr_name is None or not hasattr(obj.__class__, attr_name):
        return None, None, ()
    class_attr = getattr(obj.__class__, attr_name)
    if isinstance(class_attr, property):
        if not class_attr.fget:
            raise ValueError(f"Attribute {attr_name} has no getter.")
        hint = get_type_hints(class_attr.fget)['return']
    else:
        try:
            hint = get_type_hints(obj.__class__)[attr_name]
        except KeyError:
            hint = type(class_attr)
    origin = get_origin(hint)
    args = get_args(hint)
    if origin is Mapped:
        return args[0], get_origin(args[0]), get_args(args[0])
    else:
        return hint, origin, args


def table_rows_as_str(row_dict: Dict[str, Any], columns_per_row: int = 9):
    """
    Print a table row.

    :param row_dict: The row to print.
    :param columns_per_row: The maximum number of columns per row.
    """
    all_items = list(row_dict.items())
    # make items a list of n rows such that each row has a max size of 4
    all_items = [all_items[i:i + columns_per_row] for i in range(0, len(all_items), columns_per_row)]
    keys = [list(map(lambda i: i[0], row)) for row in all_items]
    values = [list(map(lambda i: i[1], row)) for row in all_items]
    all_table_rows = []
    for row_keys, row_values in zip(keys, values):
        table = tabulate([row_values], headers=row_keys, tablefmt='plain')
        all_table_rows.append(table)
    return "\n".join(all_table_rows)


def row_to_dict(obj):
    return {
        col.name: getattr(obj, col.name)
        for col in obj.__table__.columns
        if not col.primary_key and not col.foreign_keys
    }


def get_attribute_name(obj: Any, attribute: Optional[Any] = None, attribute_type: Optional[Type] = None,
                       possible_value: Optional[Any] = None) -> Optional[str]:
    """
    Get the name of an attribute from an object. The attribute can be given as a value, a type or a target value.
    And this method will try to find the attribute name using the given information.

    :param obj: The object to get the attribute name from.
    :param attribute: The attribute to get the name of.
    :param attribute_type: The type of the attribute to get the name of.
    :param possible_value: A possible value of the attribute to get the name of.
    :return: The name of the attribute.
    """
    attribute_name: Optional[str] = None
    if attribute_name is None and attribute is not None:
        attribute_name = get_attribute_name_from_value(obj, attribute)
    if attribute_name is None and attribute_type is not None:
        attribute_name = get_attribute_by_type(obj, attribute_type)[0]
    if attribute_name is None and possible_value is not None:
        attribute_name = get_attribute_by_type(obj, type(possible_value))[0]
    return attribute_name


def get_attribute_by_type(obj: Any, prop_type: Type) -> Tuple[Optional[str], Optional[Any]]:
    """
    Get a property from an object by type.

    :param obj: The object to get the property from.
    :param prop_type: The type of the property.
    """
    for name in dir(obj):
        if name.startswith("_") or callable(getattr(obj, name)):
            continue
        if isinstance(getattr(obj, name), (MetaData, registry)):
            continue
        prop_value = getattr(obj, name)
        if isinstance(prop_value, prop_type):
            return name, prop_value
        if hasattr(prop_value, "__iter__") and not isinstance(prop_value, str):
            if len(prop_value) > 0 and any(isinstance(v, prop_type) for v in prop_value):
                return name, prop_value
            else:
                # get args of type hint
                hint, origin, args = get_hint_for_attribute(name, obj)
                if origin in [list, set, tuple, dict, List, Set, Tuple, Dict]:
                    if prop_type is args[0]:
                        return name, prop_value
        else:
            # get the type hint of the attribute
            hint, origin, args = get_hint_for_attribute(name, obj)
            if hint is prop_type:
                return name, prop_value
            elif origin in [list, set, tuple, dict, List, Set, Tuple, Dict]:
                if prop_type is args[0]:
                    return name, prop_value
    return None, None


def get_attribute_name_from_value(obj: Any, attribute_value: Any) -> Optional[str]:
    """
    Get the name of an attribute from an object.

    :param obj: The object to get the attribute name from.
    :param attribute_value: The attribute value to get the name of.
    """
    for name in dir(obj):
        if name.startswith("_") or callable(getattr(obj, name)):
            continue
        prop_value = getattr(obj, name)
        if prop_value is attribute_value:
            return name


def get_attribute_values_transitively(obj: Any, attribute: Any) -> Any:
    """
    Get an attribute from a python object, if it is iterable, get the attribute values from all elements and unpack them
    into a list.

    :param obj: The object to get the sub attribute from.
    :param attribute: The  attribute to get.
    """
    if hasattr(obj, "__iter__") and not isinstance(obj, str):
        if isinstance(obj, (dict, UserDict)):
            all_values = [get_attribute_values_transitively(v, attribute) for v in obj.values()
                          if not isinstance(v, (str, type)) and hasattr(v, attribute)]
        else:
            all_values = [get_attribute_values_transitively(a, attribute) for a in obj
                          if not isinstance(a, (str, type)) and hasattr(a, attribute)]
        if can_be_a_set(all_values):
            return set().union(*all_values)
        else:
            return set(all_values)
    return getattr(obj, attribute)


def can_be_a_set(value: Any) -> bool:
    """
    Check if a value can be a set.

    :param value: The value to check.
    """
    if hasattr(value, "__iter__") and not isinstance(value, str):
        if len(value) > 0 and any(hasattr(v, "__iter__") and not isinstance(v, str) for v in value):
            return False
        else:
            return True
    else:
        return False


def get_all_subclasses(cls: Type) -> Dict[str, Type]:
    """
    Get all subclasses of a class recursively.

    :param cls: The class to get the subclasses of.
    :return: A dictionary of all subclasses.
    """
    all_subclasses: Dict[str, Type] = {}
    for sub_cls in cls.__subclasses__():
        all_subclasses[sub_cls.__name__.lower()] = sub_cls
        all_subclasses.update(get_all_subclasses(sub_cls))
    return all_subclasses


def make_set(value: Any) -> Set:
    """
    Make a set from a value.

    :param value: The value to make a set from.
    """
    return set(value) if is_iterable(value) else {value}


def make_value_or_raise_error(value: Any) -> Any:
    """
    Make a value or raise an error if the value is not a single value.

    :param value: The value to check.
    """
    if hasattr(value, "__iter__") and not isinstance(value, str):
        if hasattr(value, "__len__") and len(value) == 1:
            return list(value)[0]
        else:
            raise ValueError(f"Expected a single value, got {value}")
    return value


def tree_to_graph(root_node: Node) -> nx.DiGraph:
    """
    Convert anytree to a networkx graph.

    :param root_node: The root node of the tree.
    :return: A networkx graph.
    """
    graph = nx.DiGraph()
    unique_node_names = get_unique_node_names_func(root_node)

    def add_edges(node):
        if unique_node_names(node) not in graph.nodes:
            graph.add_node(unique_node_names(node))
        for child in node.children:
            if unique_node_names(child) not in graph.nodes:
                graph.add_node(unique_node_names(child))
            graph.add_edge(unique_node_names(node), unique_node_names(child), weight=child.weight)
            add_edges(child)

    add_edges(root_node)
    return graph


def get_unique_node_names_func(root_node) -> Callable[[Node], str]:
    nodes = [root_node]

    def get_all_nodes(node):
        for c in node.children:
            nodes.append(c)
            get_all_nodes(c)

    get_all_nodes(root_node)

    def nodenamefunc(node: Node):
        """
        Set the node name for the dot exporter.
        """
        similar_nodes = [n for n in nodes if n.name == node.name]
        node_idx = similar_nodes.index(node)
        return node.name if node_idx == 0 else f"{node.name}_{node_idx}"

    return nodenamefunc


def edge_attr_setter(parent, child):
    """
    Set the edge attributes for the dot exporter.
    """
    if child and hasattr(child, "weight") and child.weight:
        return f'style="bold", label=" {child.weight}"'
    return ""


def render_tree(root: Node, use_dot_exporter: bool = False,
                filename: str = "scrdr"):
    """
    Render the tree using the console and optionally export it to a dot file.

    :param root: The root node of the tree.
    :param use_dot_exporter: Whether to export the tree to a dot file.
    :param filename: The name of the file to export the tree to.
    """
    if not root:
        logging.warning("No rules to render")
        return
    for pre, _, node in RenderTree(root):
        print(f"{pre}{node.weight if hasattr(node, 'weight') and node.weight else ''} {node.__str__()}")
    if use_dot_exporter:
        unique_node_names = get_unique_node_names_func(root)

        de = DotExporter(root,
                         nodenamefunc=unique_node_names,
                         edgeattrfunc=edge_attr_setter
                         )
        de.to_dotfile(f"{filename}{'.dot'}")
        de.to_picture(f"{filename}{'.png'}")


def draw_tree(root: Node, fig: plt.Figure):
    """
    Draw the tree using matplotlib and networkx.
    """
    if root is None:
        return
    fig.clf()
    graph = tree_to_graph(root)
    fig_sz_x = 13
    fig_sz_y = 10
    fig.set_size_inches(fig_sz_x, fig_sz_y)
    pos = nx.drawing.nx_agraph.graphviz_layout(graph, prog="dot")
    # scale down pos
    max_pos_x = max([v[0] for v in pos.values()])
    max_pos_y = max([v[1] for v in pos.values()])
    pos = {k: (v[0] * fig_sz_x / max_pos_x, v[1] * fig_sz_y / max_pos_y) for k, v in pos.items()}
    nx.draw(graph, pos, with_labels=True, node_color="lightblue", edge_color="gray", node_size=1000,
            ax=fig.gca(), node_shape="o", font_size=8)
    nx.draw_networkx_edge_labels(graph, pos, edge_labels=nx.get_edge_attributes(graph, 'weight'),
                                 ax=fig.gca(), rotate=False, clip_on=False)
    plt.pause(0.1)



from __future__ import annotations

import ast
import logging
from _ast import AST

from sqlalchemy.orm import Session
from typing_extensions import Type, Optional, Any, List, Union, Tuple, Dict, Set

from .case import create_case, Case
from ..utils import SubclassJSONSerializer, get_full_class_name, get_type_from_string


class VariableVisitor(ast.NodeVisitor):
    """
    A visitor to extract all variables and comparisons from a python expression represented as an AST tree.
    """
    compares: List[Tuple[Union[ast.Name, ast.Call], ast.cmpop, Union[ast.Name, ast.Call]]]
    variables: Set[str]
    all: List[ast.BoolOp]

    def __init__(self):
        self.variables = set()
        self.attributes: Dict[ast.Name, ast.Attribute] = {}
        self.compares = list()
        self.binary_ops = list()
        self.all = list()

    def visit_Attribute(self, node):
        self.all.append(node)
        self.attributes[node.value] = node
        self.generic_visit(node)

    def visit_BinOp(self, node):
        self.binary_ops.append(node)
        self.all.append(node)
        self.generic_visit(node)

    def visit_BoolOp(self, node):
        self.all.append(node)
        self.generic_visit(node)

    def visit_Compare(self, node):
        self.all.append(node)
        self.compares.append([node.left, node.ops[0], node.comparators[0]])
        self.generic_visit(node)

    def visit_Name(self, node):
        if f"__{node.id}__" not in dir(__builtins__) and node not in self.attributes:
            self.variables.add(node.id)
        self.generic_visit(node)


class CallableExpression(SubclassJSONSerializer):
    """
    A callable that is constructed from a string statement written by an expert.
    """
    conclusion_type: Type
    """
    The type of the output of the callable, used for assertion.
    """
    expression_tree: AST
    """
    The AST tree parsed from the user input.
    """
    user_input: str
    """
    The input given by the expert.
    """
    session: Optional[Session]
    """
    The sqlalchemy orm session.
    """
    visitor: VariableVisitor
    """
    A visitor to extract all variables and comparisons from a python expression represented as an AST tree.
    """
    code: Any
    """
    The code that was compiled from the expression tree
    """
    compares_column_offset: List[int]
    """
    The start and end indices of each comparison in the string of user input.
    """

    def __init__(self, user_input: str, conclusion_type: Optional[Type] = None, expression_tree: Optional[AST] = None,
                 session: Optional[Session] = None):
        """
        Create a callable expression.

        :param user_input: The input given by the expert.
        :param conclusion_type: The type of the output of the callable.
        :param expression_tree: The AST tree parsed from the user input.
        :param session: The sqlalchemy orm session.
        """
        self.session = session
        self.user_input: str = user_input
        self.parsed_user_input = self.parse_user_input(user_input, session)
        self.conclusion_type = conclusion_type
        self.update_expression(self.parsed_user_input, expression_tree)

    @staticmethod
    def parse_user_input(user_input: str, session: Optional[Session] = None) -> str:
        if ',' in user_input:
            user_input = user_input.split(',')
            user_input = [f"({u.strip()})" for u in user_input]
            user_input = ' & '.join(user_input) if session else ' and '.join(user_input)
        elif session:
            user_input = user_input.replace(" and ", " & ")
            user_input = user_input.replace(" or ", " | ")
        return user_input

    def update_expression(self, user_input: str, expression_tree: Optional[AST] = None):
        if not expression_tree:
            expression_tree = parse_string_to_expression(user_input)
        self.expression_tree: AST = expression_tree
        self.visitor = VariableVisitor()
        self.visitor.visit(expression_tree)
        variables_str = self.visitor.variables
        attributes_str = get_attributes_str(self.visitor)
        for v in variables_str | attributes_str:
            if not v.startswith("case."):
                self.parsed_user_input = self.parsed_user_input.replace(v, f"case.{v}")
        self.expression_tree = parse_string_to_expression(self.parsed_user_input)
        self.compares_column_offset = [(c[0].col_offset, c[2].end_col_offset) for c in self.visitor.compares]
        self.code = compile_expression_to_code(self.expression_tree)

    def __call__(self, case: Any, **kwargs) -> Any:
        try:
            if not isinstance(case, Case):
                case = create_case(case, max_recursion_idx=3)
            output = eval(self.code)
            if self.conclusion_type:
                assert isinstance(output, self.conclusion_type), (f"Expected output type {self.conclusion_type},"
                                                                  f" got {type(output)}")
            return output
        except Exception as e:
            raise ValueError(f"Error during evaluation: {e}")

    def combine_with(self, other: 'CallableExpression') -> 'CallableExpression':
        """
        Combine this callable expression with another callable expression using the 'and' operator.
        """
        new_user_input = f"({self.user_input}) and ({other.user_input})"
        return CallableExpression(new_user_input, conclusion_type=self.conclusion_type, session=self.session)

    def __str__(self):
        """
        Return the user string where each compare is written in a line using compare column offset start and end.
        """
        user_input = self.parsed_user_input
        binary_ops = sorted(self.visitor.binary_ops, key=lambda x: x.end_col_offset)
        binary_ops_indices = [b.end_col_offset for b in binary_ops]
        all_binary_ops = []
        prev_e = 0
        for i, e in enumerate(binary_ops_indices):
            if i == 0:
                all_binary_ops.append(user_input[:e])
            else:
                all_binary_ops.append(user_input[prev_e:e])
            prev_e = e
        return "\n".join(all_binary_ops) if len(all_binary_ops) > 0 else user_input

    def _to_json(self) -> Dict[str, Any]:
        return {"user_input": self.user_input, "conclusion_type": get_full_class_name(self.conclusion_type)}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> CallableExpression:
        return cls(user_input=data["user_input"], conclusion_type=get_type_from_string(data["conclusion_type"]))


def compile_expression_to_code(expression_tree: AST) -> Any:
    """
    Compile an expression tree that was parsed from string into code that can be executed using 'eval(code)'

    :param expression_tree: The parsed expression tree.
    :return: The code that was compiled from the expression tree.
    """
    return compile(expression_tree, filename="<string>", mode="eval")


def assert_context_contains_needed_information(case: Any, context: Dict[str, Any],
                                               visitor: VariableVisitor) -> Tuple[Set[str], Set[str]]:
    """
    Asserts that the variables mentioned in the expression visited by visitor are all in the given context.

    :param case: The case to check the context for.
    :param context: The context to check.
    :param visitor: The visitor that visited the expression.
    :return: The found variables and attributes.
    """
    found_variables = set()
    for key in visitor.variables:
        if key not in context:
            raise ValueError(f"Variable {key} not found in the case {case}")
        found_variables.add(key)

    found_attributes = get_attributes_str(visitor)
    for attr in found_attributes:
        if attr not in context:
            raise ValueError(f"Attribute {attr} not found in the case {case}")
    return found_variables, found_attributes


def get_attributes_str(visitor: VariableVisitor) -> Set[str]:
    """
    Get the string representation of the attributes in the given visitor.

    :param visitor: The visitor that visited the expression.
    :return: The string representation of the attributes.
    """
    found_attributes = set()
    for key, ast_attr in visitor.attributes.items():
        str_attr = ""
        while isinstance(key, ast.Attribute):
            if len(str_attr) > 0:
                str_attr = f"{key.attr}.{str_attr}"
            else:
                str_attr = key.attr
            key = key.value
        str_attr = f"{key.id}.{str_attr}" if len(str_attr) > 0 else f"{key.id}.{ast_attr.attr}"
        found_attributes.add(str_attr)
    return found_attributes


def parse_string_to_expression(expression_str: str) -> AST:
    """
    Parse a string statement into an AST expression.

    :param expression_str: The string which will be parsed.
    :return: The parsed expression.
    """
    tree = ast.parse(expression_str, mode='eval')
    logging.debug(f"AST parsed successfully: {ast.dump(tree)}")
    return tree





class InvalidOperator(Exception):
    def __init__(self, rule_str: str, valid_operators: list):
        self.rule_str = rule_str
        self.valid_operators = valid_operators

    def __str__(self):
        return f"Invalid operator in {self.rule_str}, valid operators are: {self.valid_operators}"




from __future__ import annotations

import os
import pickle

import sqlalchemy
from sqlalchemy import ForeignKey
from sqlalchemy.orm import MappedAsDataclass, Mapped, mapped_column, relationship
from typing_extensions import Tuple, List, Set, Optional
from ucimlrepo import fetch_ucirepo

from .datastructures import Case, create_cases_from_dataframe, Category, CaseAttribute


def load_cached_dataset(cache_file):
    """Loads the dataset from cache if it exists."""
    dataset = {}
    for key in ["features", "targets", "ids"]:
        part_file = cache_file.replace(".pkl", f"_{key}.pkl")
        if not os.path.exists(part_file):
            return None
        with open(part_file, "rb") as f:
            dataset[key] = pickle.load(f)
    return dataset


def save_dataset_to_cache(dataset, cache_file):
    """Saves only essential parts of the dataset to cache."""
    dataset_to_cache = {
        "features": dataset.data.features,
        "targets": dataset.data.targets,
        "ids": dataset.data.ids,
    }

    for key, value in dataset_to_cache.items():
        with open(cache_file.replace(".pkl", f"_{key}.pkl"), "wb") as f:
            pickle.dump(dataset_to_cache[key], f)
    print("Dataset cached successfully.")


def get_dataset(dataset_id, cache_file: Optional[str] = None):
    """Fetches dataset from cache or downloads it if not available."""
    dataset = load_cached_dataset(cache_file) if cache_file else None
    if dataset is None:
        print("Downloading dataset...")
        dataset = fetch_ucirepo(id=dataset_id)

        # Check if dataset is valid before caching
        if dataset is None or not hasattr(dataset, "data"):
            print("Error: Failed to fetch dataset.")
            return None

        if cache_file:
            save_dataset_to_cache(dataset, cache_file)

        dataset = {
            "features": dataset.data.features,
            "targets": dataset.data.targets,
            "ids": dataset.data.ids,
        }

    return dataset


def load_zoo_dataset(cache_file: Optional[str] = None) -> Tuple[List[Case], List[Species]]:
    """
    Load the zoo dataset.

    :param cache_file: the cache file to store the dataset or load it from.
    :return: all cases and targets.
    """
    # fetch dataset
    zoo = get_dataset(111, cache_file)

    # data (as pandas dataframes)
    X = zoo['features']
    y = zoo['targets']
    # get ids as list of strings
    ids = zoo['ids'].values.flatten()
    all_cases = create_cases_from_dataframe(X)

    category_names = ["mammal", "bird", "reptile", "fish", "amphibian", "insect", "molusc"]
    category_id_to_name = {i + 1: name for i, name in enumerate(category_names)}
    # targets = [getattr(SpeciesCol, category_id_to_name[i]) for i in y.values.flatten()]
    targets = [Species.from_str(category_id_to_name[i]) for i in y.values.flatten()]
    return all_cases, targets


class Species(Category):
    mammal = "mammal"
    bird = "bird"
    reptile = "reptile"
    fish = "fish"
    amphibian = "amphibian"
    insect = "insect"
    molusc = "molusc"


class Habitat(Category):
    """
    A habitat category is a category that represents the habitat of an animal.
    """
    land = "land"
    water = "water"
    air = "air"


# SpeciesCol = Column.create_from_enum(Species, mutually_exclusive=True)
# HabitatCol = Column.create_from_enum(Habitat, mutually_exclusive=False)


class Base(sqlalchemy.orm.DeclarativeBase):
    pass


class HabitatTable(MappedAsDataclass, Base):
    __tablename__ = "Habitat"

    id: Mapped[int] = mapped_column(init=False, primary_key=True, autoincrement=True)
    habitat: Mapped[Habitat]
    animal_id = mapped_column(ForeignKey("Animal.id"), init=False)

    def __hash__(self):
        return hash(self.habitat)

    def __str__(self):
        return self.habitat.value

    def __repr__(self):
        return self.__str__()


class Animal(MappedAsDataclass, Base):
    __tablename__ = "Animal"

    id: Mapped[int] = mapped_column(init=False, primary_key=True, autoincrement=True)
    name: Mapped[str]
    hair: Mapped[bool]
    feathers: Mapped[bool]
    eggs: Mapped[bool]
    milk: Mapped[bool]
    airborne: Mapped[bool]
    aquatic: Mapped[bool]
    predator: Mapped[bool]
    toothed: Mapped[bool]
    backbone: Mapped[bool]
    breathes: Mapped[bool]
    venomous: Mapped[bool]
    fins: Mapped[bool]
    legs: Mapped[int]
    tail: Mapped[bool]
    domestic: Mapped[bool]
    catsize: Mapped[bool]
    species: Mapped[Species] = mapped_column(nullable=True)

    habitats: Mapped[Set[HabitatTable]] = relationship(default_factory=set)



from __future__ import annotations

import json
from abc import ABC, abstractmethod

from sqlalchemy.orm import DeclarativeBase as SQLTable, MappedColumn as SQLColumn, Session
from typing_extensions import Optional, Dict, TYPE_CHECKING, List, Tuple, Type, Union, Any

from .datastructures import (Case, PromptFor, CallableExpression, CaseAttribute, CaseQuery)
from .datastructures.case import show_current_and_corner_cases
from .prompt import prompt_user_for_expression, prompt_user_about_case
from .utils import get_all_subclasses, is_iterable

if TYPE_CHECKING:
    from .rdr import Rule


class Expert(ABC):
    """
    The Abstract Expert class, all experts should inherit from this class.
    An expert is a class that can provide differentiating features and conclusions for a case when asked.
    The expert can compare a case with a corner case and provide the differentiating features and can also
    provide one or multiple conclusions for a case.
    """
    all_expert_answers: Optional[List] = None
    """
    A list of all expert answers, used for testing purposes.
    """
    use_loaded_answers: bool = False
    """
    A flag to indicate if the expert should use loaded answers or not.
    """
    known_categories: Optional[Dict[str, Type[CaseAttribute]]] = None
    """
    The known categories (i.e. Column types) to use.
    """

    @abstractmethod
    def ask_for_conditions(self, x: Case, targets: List[CaseAttribute], last_evaluated_rule: Optional[Rule] = None) \
            -> CallableExpression:
        """
        Ask the expert to provide the differentiating features between two cases or unique features for a case
        that doesn't have a corner case to compare to.

        :param x: The case to classify.
        :param targets: The target categories to compare the case with.
        :param last_evaluated_rule: The last evaluated rule.
        :return: The differentiating features as new rule conditions.
        """
        pass

    @abstractmethod
    def ask_for_extra_conclusions(self, x: Case, current_conclusions: List[CaseAttribute]) \
            -> Dict[CaseAttribute, CallableExpression]:
        """
        Ask the expert to provide extra conclusions for a case by providing a pair of category and conditions for
        that category.

        :param x: The case to classify.
        :param current_conclusions: The current conclusions for the case.
        :return: The extra conclusions for the case.
        """
        pass

    @abstractmethod
    def ask_if_conclusion_is_correct(self, x: Case, conclusion: CaseAttribute,
                                     targets: Optional[List[CaseAttribute]] = None,
                                     current_conclusions: Optional[List[CaseAttribute]] = None) -> bool:
        """
        Ask the expert if the conclusion is correct.

        :param x: The case to classify.
        :param conclusion: The conclusion to check.
        :param targets: The target categories to compare the case with.
        :param current_conclusions: The current conclusions for the case.
        """
        pass

    def ask_for_conclusion(self, case_query: CaseQuery,
                           session: Optional[Session] = None) -> Optional[CallableExpression]:
        """
        Ask the expert to provide a relational conclusion for the case.

        :param case_query: The case query containing the case to find a conclusion for.
        :param session: The sqlalchemy orm session to use if the case is a Table.
        :return: A callable expression that can be called with a new case as an argument.
        """


class Human(Expert):
    """
    The Human Expert class, an expert that asks the human to provide differentiating features and conclusions.
    """

    def __init__(self, use_loaded_answers: bool = False, session: Optional[Session] = None):
        self.all_expert_answers = []
        self.use_loaded_answers = use_loaded_answers
        self.session = session

    def save_answers(self, path: str, append: bool = False):
        """
        Save the expert answers to a file.

        :param path: The path to save the answers to.
        :param append: A flag to indicate if the answers should be appended to the file or not.
        """
        if append:
            # read the file and append the new answers
            with open(path + '.json', "r") as f:
                all_answers = json.load(f)
                all_answers.extend(self.all_expert_answers)
            with open(path + '.json', "w") as f:
                json.dump(all_answers, f)
        else:
            with open(path + '.json', "w") as f:
                json.dump(self.all_expert_answers, f)

    def load_answers(self, path: str):
        """
        Load the expert answers from a file.

        :param path: The path to load the answers from.
        """
        with open(path + '.json', "r") as f:
            self.all_expert_answers = json.load(f)

    def ask_for_conditions(self, case: Case,
                           targets: Union[List[CaseAttribute], List[SQLColumn]],
                           last_evaluated_rule: Optional[Rule] = None) \
            -> CallableExpression:
        if not self.use_loaded_answers:
            show_current_and_corner_cases(case, targets, last_evaluated_rule=last_evaluated_rule)
        return self._get_conditions(case, targets)

    def _get_conditions(self, case: Case, targets: List[CaseAttribute]) \
            -> CallableExpression:
        """
        Ask the expert to provide the differentiating features between two cases or unique features for a case
        that doesn't have a corner case to compare to.

        :param case: The case to classify.
        :param targets: The target categories to compare the case with.
        :return: The differentiating features as new rule conditions.
        """
        targets = targets if isinstance(targets, list) else [targets]
        condition = None
        for target in targets:
            target_name = target.__class__.__name__
            user_input = None
            if self.use_loaded_answers:
                user_input = self.all_expert_answers.pop(0)
            if user_input:
                condition = CallableExpression(user_input, bool, session=self.session)
            else:
                user_input, condition = prompt_user_for_expression(case, PromptFor.Conditions, target_name, bool)
            if not self.use_loaded_answers:
                self.all_expert_answers.append(user_input)
        return condition

    def ask_for_extra_conclusions(self, case: Case, current_conclusions: List[CaseAttribute]) \
            -> Dict[CaseAttribute, CallableExpression]:
        """
        Ask the expert to provide extra conclusions for a case by providing a pair of category and conditions for
        that category.

        :param case: The case to classify.
        :param current_conclusions: The current conclusions for the case.
        :return: The extra conclusions for the case.
        """
        extra_conclusions = {}
        while True:
            category = self.ask_for_conclusion(CaseQuery(case), current_conclusions)
            if not category:
                break
            extra_conclusions[category] = self._get_conditions(case, category)
        return extra_conclusions

    def ask_for_conclusion(self, case_query: CaseQuery,
                           current_conclusions: Optional[List[Any]] = None)\
            -> Optional[CallableExpression]:
        """
        Ask the expert to provide a conclusion for the case.

        :param case_query: The case query containing the case to find a conclusion for.
        :param current_conclusions: The current conclusions for the case if any.
        :return: The conclusion for the case.
        """
        case = case_query.case
        attribute_name = case_query.attribute_name
        attribute_type = case_query.attribute_type
        if self.use_loaded_answers:
            expert_input = self.all_expert_answers.pop(0)
            expression = CallableExpression(expert_input, conclusion_type=attribute_type, session=self.session)
        else:
            show_current_and_corner_cases(case, current_conclusions=current_conclusions)
            expert_input, expression = prompt_user_for_expression(case, PromptFor.Conclusion, attribute_name,
                                                                  attribute_type)
            self.all_expert_answers.append(expert_input)
        return expression

    def get_category_type(self, cat_name: str) -> Optional[Type[CaseAttribute]]:
        """
        Get the category type from the known categories.

        :param cat_name: The name of the category.
        :return: The category type.
        """
        cat_name = cat_name.lower()
        self.known_categories = get_all_subclasses(CaseAttribute) if not self.known_categories else self.known_categories
        self.known_categories.update(CaseAttribute.registry)
        category_type = None
        if cat_name in self.known_categories:
            category_type = self.known_categories[cat_name]
        return category_type

    def ask_if_category_is_mutually_exclusive(self, category_name: str) -> bool:
        """
        Ask the expert if the new category can have multiple values.

        :param category_name: The name of the category to ask about.
        """
        question = f"Can a case have multiple values of the new category {category_name}? (y/n):"
        return not self.ask_yes_no_question(question)

    def ask_if_conclusion_is_correct(self, x: Case, conclusion: CaseAttribute,
                                     targets: Optional[List[CaseAttribute]] = None,
                                     current_conclusions: Optional[List[CaseAttribute]] = None) -> bool:
        """
        Ask the expert if the conclusion is correct.

        :param x: The case to classify.
        :param conclusion: The conclusion to check.
        :param targets: The target categories to compare the case with.
        :param current_conclusions: The current conclusions for the case.
        """
        question = ""
        if not self.use_loaded_answers:
            targets = targets or []
            targets = targets if isinstance(targets, list) else [targets]
            x.conclusions = current_conclusions
            x.targets = targets
            question = f"Is the conclusion {conclusion} correct for the case (y/n):" \
                       f"\n{str(x)}"
        return self.ask_yes_no_question(question)

    def ask_yes_no_question(self, question: str) -> bool:
        """
        Ask the expert a yes or no question.

        :param question: The question to ask.
        :return: The answer to the question.
        """
        if not self.use_loaded_answers:
            print(question)
        while True:
            if self.use_loaded_answers:
                answer = self.all_expert_answers.pop(0)
            else:
                answer = input()
                self.all_expert_answers.append(answer)
            if answer.lower() == "y":
                return True
            elif answer.lower() == "n":
                return False



from __future__ import annotations

from collections import UserDict
from dataclasses import dataclass
from enum import Enum

from pandas import DataFrame
from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase as SQLTable, MappedColumn as SQLColumn, registry
from typing_extensions import Any, Optional, Dict, Type, Set, Hashable, Union, List, TYPE_CHECKING

from ..utils import make_set, row_to_dict, table_rows_as_str, get_value_type_from_type_hint, SubclassJSONSerializer

if TYPE_CHECKING:
    from ripple_down_rules.rules import Rule
    from .callable_expression import CallableExpression


class Case(UserDict, SubclassJSONSerializer):
    """
    A collection of attributes that represents a set of constraints on a case. This is a dictionary where the keys are
    the names of the attributes and the values are the attributes. All are stored in lower case.
    """

    def __init__(self, _id: Optional[Hashable] = None, _type: Optional[Type] = None, **kwargs):
        """
        Create a new row.

        :param _id: The id of the row.
        :param kwargs: The attributes of the row.
        """
        super().__init__(kwargs)
        self._id = _id if _id else id(self)
        self._type = _type

    @classmethod
    def from_obj(cls, obj: Any, obj_name: Optional[str] = None, max_recursion_idx: int = 3) -> Case:
        """
        Create a row from an object.

        :param obj: The object to create a row from.
        :param max_recursion_idx: The maximum recursion index to prevent infinite recursion.
        :param obj_name: The name of the object.
        :return: The row of the object.
        """
        return create_case(obj, max_recursion_idx=max_recursion_idx, obj_name=obj_name)

    def __getitem__(self, item: str) -> Any:
        return super().__getitem__(item.lower())

    def __setitem__(self, name: str, value: Any):
        name = name.lower()
        if name in self:
            if isinstance(self[name], set):
                self[name].update(make_set(value))
            elif isinstance(value, set):
                value.update(make_set(self[name]))
                super().__setitem__(name, value)
            else:
                super().__setitem__(name, make_set([self[name], value]))
        else:
            super().__setitem__(name, value)
        setattr(self, name, self[name])

    def __contains__(self, item):
        if isinstance(item, (type, Enum)):
            item = item.__name__
        return super().__contains__(item.lower())

    def __delitem__(self, key):
        super().__delitem__(key.lower())

    def __hash__(self):
        return self._id

    def _to_json(self) -> Dict[str, Any]:
        serializable = {k: v for k, v in self.items() if not k.startswith("_")}
        serializable["_id"] = self._id
        return {k: v.to_json() if isinstance(v, SubclassJSONSerializer) else v for k, v in serializable.items()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Case:
        id_ = data.pop("_id")
        return cls(_id=id_, **data)


@dataclass
class CaseAttributeValue(SubclassJSONSerializer):
    """
    A column value is a value in a column.
    """
    id: Hashable
    """
    The row id of the column value.
    """
    value: Any
    """
    The value of the column.
    """

    def __eq__(self, other):
        if not isinstance(other, CaseAttributeValue):
            return False
        return self.value == other.value

    def __hash__(self):
        return self.id

    def _to_json(self) -> Dict[str, Any]:
        return {"id": self.id, "value": self.value}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> CaseAttributeValue:
        return cls(id=data["id"], value=data["value"])


class CaseAttribute(set, SubclassJSONSerializer):
    nullable: bool = True
    """
    A boolean indicating whether the case attribute can be None or not.
    """
    mutually_exclusive: bool = False
    """
    A boolean indicating whether the case attribute is mutually exclusive or not. (i.e. can only have one value)
    """

    def __init__(self, values: Set[CaseAttributeValue]):
        """
        Create a new case attribute.

        :param values: The values of the case attribute.
        """
        values = self._type_cast_values_to_set_of_case_attribute_values(values)
        self.id_value_map: Dict[Hashable, Union[CaseAttributeValue, Set[CaseAttributeValue]]] = {id(v): v for v in values}
        super().__init__([v.value for v in values])

    @staticmethod
    def _type_cast_values_to_set_of_case_attribute_values(values: Set[Any]) -> Set[CaseAttributeValue]:
        """
        Type cast values to a set of case attribute values.

        :param values: The values to type cast.
        """
        values = make_set(values)
        if len(values) > 0 and not isinstance(next(iter(values)), CaseAttributeValue):
            values = {CaseAttributeValue(id(values), v) for v in values}
        return values

    @classmethod
    def from_obj(cls, values: Set[Any], row_obj: Optional[Any] = None) -> CaseAttribute:
        id_ = id(row_obj) if row_obj else id(values)
        values = make_set(values)
        return cls({CaseAttributeValue(id_, v) for v in values})

    @property
    def as_dict(self) -> Dict[str, Any]:
        """
        Get the case attribute as a dictionary.

        :return: The case attribute as a dictionary.
        """
        return {self.__class__.__name__: self}

    def filter_by(self, condition: CallableExpression) -> CaseAttribute:
        """
        Filter the column by a condition.

        :param condition: The condition to filter by.
        :return: The filtered column.
        """
        return self.__class__({v for v in self if condition(v)})

    def __eq__(self, other):
        if not isinstance(other, set):
            return super().__eq__(make_set(other))
        return super().__eq__(other)

    def __hash__(self):
        return hash(tuple(self.id_value_map.values()))

    def __str__(self):
        if len(self) == 0:
            return "None"
        return str({v for v in self}) if len(self) > 1 else str(next(iter(self)))

    def _to_json(self) -> Dict[str, Any]:
        return {id_: v.to_json() if isinstance(v, SubclassJSONSerializer) else v
                for id_, v in self.id_value_map.items()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> CaseAttribute:
        return cls({CaseAttributeValue.from_json(v) for id_, v in data.items()})


def create_cases_from_dataframe(df: DataFrame) -> List[Case]:
    """
    Create cases from a pandas DataFrame.

    :param df: The DataFrame to create cases from.
    :return: The cases of the DataFrame.
    """
    cases = []
    attribute_names = list(df.columns)
    for row_id, case in df.iterrows():
        case = {col_name: case[col_name].item() for col_name in attribute_names}
        cases.append(Case(_id=row_id, _type=DataFrame, **case))
    return cases


def create_case(obj: Any, recursion_idx: int = 0, max_recursion_idx: int = 0,
                obj_name: Optional[str] = None, parent_is_iterable: bool = False) -> Case:
    """
    Create a case from an object.

    :param obj: The object to create a case from.
    :param recursion_idx: The current recursion index.
    :param max_recursion_idx: The maximum recursion index to prevent infinite recursion.
    :param obj_name: The name of the object.
    :param parent_is_iterable: Boolean indicating whether the parent object is iterable or not.
    :return: The case that represents the object.
    """
    if isinstance(obj, DataFrame):
        return create_cases_from_dataframe(obj)
    if isinstance(obj, Case):
        return obj
    if ((recursion_idx > max_recursion_idx) or (obj.__class__.__module__ == "builtins")
            or (obj.__class__ in [MetaData, registry])):
        return Case(_id=id(obj), _type=obj.__class__,
                    **{obj_name or obj.__class__.__name__: make_set(obj) if parent_is_iterable else obj})
    case = Case(_id=id(obj), _type=obj.__class__)
    for attr in dir(obj):
        if attr.startswith("_") or callable(getattr(obj, attr)):
            continue
        attr_value = getattr(obj, attr)
        case = create_or_update_case_from_attribute(attr_value, attr, obj, attr, recursion_idx,
                                                    max_recursion_idx, parent_is_iterable, case)
    return case


def create_or_update_case_from_attribute(attr_value: Any, name: str, obj: Any, obj_name: Optional[str] = None,
                                         recursion_idx: int = 0, max_recursion_idx: int = 1,
                                         parent_is_iterable: bool = False,
                                         case: Optional[Case] = None) -> Case:
    """
    Create or update a case from an attribute of the object that the case represents.

    :param attr_value: The attribute value.
    :param name: The name of the attribute.
    :param obj: The parent object of the attribute.
    :param obj_name: The parent object name.
    :param recursion_idx: The recursion index to prevent infinite recursion.
    :param max_recursion_idx: The maximum recursion index.
    :param parent_is_iterable: Boolean indicating whether the parent object is iterable or not.
    :param case: The case to update.
    :return: The updated/created case.
    """
    if case is None:
        case = Case(_id=id(obj), _type=obj.__class__)
    if isinstance(attr_value, (dict, UserDict)):
        case.update({f"{obj_name}.{k}": v for k, v in attr_value.items()})
    if hasattr(attr_value, "__iter__") and not isinstance(attr_value, str):
        column = create_case_attribute_from_iterable_attribute(attr_value, name, obj, obj_name,
                                                               recursion_idx=recursion_idx + 1,
                                                               max_recursion_idx=max_recursion_idx)
        case[obj_name] = column
    else:
        case[obj_name] = make_set(attr_value) if parent_is_iterable else attr_value
    return case


def create_case_attribute_from_iterable_attribute(attr_value: Any, name: str, obj: Any, obj_name: Optional[str] = None,
                                                  recursion_idx: int = 0,
                                                  max_recursion_idx: int = 1) -> CaseAttribute:
    """
    Get a case attribute from an iterable attribute.

    :param attr_value: The iterable attribute to get the case from.
    :param name: The name of the case.
    :param obj: The parent object of the iterable.
    :param obj_name: The parent object name.
    :param recursion_idx: The recursion index to prevent infinite recursion.
    :param max_recursion_idx: The maximum recursion index.
    :return: A case attribute that represents the original iterable attribute.
    """
    values = attr_value.values() if isinstance(attr_value, (dict, UserDict)) else attr_value
    _type = type(list(values)[0]) if len(values) > 0 else get_value_type_from_type_hint(name, obj)
    attr_case = Case(_id=id(attr_value), _type=_type)
    case_attr = CaseAttribute.from_obj(values, row_obj=obj)
    for idx, val in enumerate(values):
        sub_attr_case = create_case(val, recursion_idx=recursion_idx,
                                    max_recursion_idx=max_recursion_idx,
                                    obj_name=obj_name, parent_is_iterable=True)
        attr_case.update(sub_attr_case)
    for sub_attr, val in attr_case.items():
        setattr(case_attr, sub_attr, val)
    return case_attr


def show_current_and_corner_cases(case: Any, targets: Optional[Union[List[CaseAttribute], List[SQLColumn]]] = None,
                                  current_conclusions: Optional[Union[List[CaseAttribute], List[SQLColumn]]] = None,
                                  last_evaluated_rule: Optional[Rule] = None) -> None:
    """
    Show the data of the new case and if last evaluated rule exists also show that of the corner case.

    :param case: The new case.
    :param targets: The target attribute of the case.
    :param current_conclusions: The current conclusions of the case.
    :param last_evaluated_rule: The last evaluated rule in the RDR.
    """
    corner_case = None
    if targets:
        targets = targets if isinstance(targets, list) else [targets]
    if current_conclusions:
        current_conclusions = current_conclusions if isinstance(current_conclusions, list) else [current_conclusions]
    targets = {f"target_{t.__class__.__name__}": t for t in targets} if targets else {}
    current_conclusions = {c.__class__.__name__: c for c in current_conclusions} if current_conclusions else {}
    if last_evaluated_rule:
        action = "Refinement" if last_evaluated_rule.fired else "Alternative"
        print(f"{action} needed for rule: {last_evaluated_rule}\n")
        corner_case = last_evaluated_rule.corner_case

    corner_row_dict = None
    if isinstance(case, SQLTable):
        case_dict = row_to_dict(case)
        if last_evaluated_rule and last_evaluated_rule.fired:
            corner_row_dict = row_to_dict(last_evaluated_rule.corner_case)
    else:
        case_dict = case
        if last_evaluated_rule and last_evaluated_rule.fired:
            corner_row_dict = corner_case

    if corner_row_dict:
        corner_conclusion = last_evaluated_rule.conclusion
        corner_row_dict.update({corner_conclusion.__class__.__name__: corner_conclusion})
        print(table_rows_as_str(corner_row_dict))
    print("=" * 50)
    case_dict.update(targets)
    case_dict.update(current_conclusions)
    print(table_rows_as_str(case_dict))



import json
from unittest import TestCase

from typing_extensions import List

from ripple_down_rules.datasets import load_zoo_dataset
from ripple_down_rules.datastructures import CaseQuery, Case
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR


class TestJSONSerialization(TestCase):
    all_cases: List[Case]
    targets: List[str]
    cache_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"

    @classmethod
    def setUpClass(cls):
        cls.all_cases, cls.targets = load_zoo_dataset(cls.cache_dir + "/zoo_dataset.pkl")

    def test_scrdr_json_serialization(self):
        scrdr = self.get_fit_scrdr()
        filename = f"{self.cache_dir}/scrdr.json"
        scrdr.save(filename)
        scrdr = SingleClassRDR.load(filename)
        for case, target in zip(self.all_cases, self.targets):
            cat = scrdr.classify(case)
            self.assertEqual(cat, target)

    def get_fit_scrdr(self, draw_tree=False) -> SingleClassRDR:
        filename = self.expert_answers_dir + "/scrdr_expert_answers_fit"
        expert = Human(use_loaded_answers=True)
        expert.load_answers(filename)

        scrdr = SingleClassRDR()
        case_queries = [CaseQuery(case, target=target) for case, target in zip(self.all_cases, self.targets)]
        scrdr.fit(case_queries, expert=expert,
                  animate_tree=draw_tree)
        for case, target in zip(self.all_cases, self.targets):
            cat = scrdr.classify(case)
            self.assertEqual(cat, target)
        return scrdr